#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
math: mathematical functions and tools for use in mystic


Functions
=========

Mystic provides a set of mathematical functions that support various
advanced optimization features such as uncertainty analysis and
parameter sensitivity.


Tools
=====

Mystic also provides a set of mathematical tools that support advanced
features such as parameter space partitioning and monte carlo estimation.
These mathematical tools are provided::

    polyeval     -- fast evaluation of an n-dimensional polynomial
    poly1d       -- generate a 1d polynomial instance
    gridpts      -- generate a set of regularly spaced points
    fillpts      -- generate a set of space-filling points
    samplepts    -- generate a set of randomly sampled points 
    tolerance    -- absolute difference plus relative difference
    almostEqual  -- test if equal within some absolute or relative tolerance
    Distribution -- generate a sampling distribution instance
"""
# functions and tools
from .poly import polyeval, poly1d
from .grid import gridpts, samplepts, fillpts
from .approx import almostEqual, tolerance


# backward compatibility
from .approx import approx_equal
from . import discrete as dirac_measure
from . import distance as paramtrans

__all__ = ['Distribution','polyeval','poly1d','gridpts','samplepts', \
           'fillpts','almostEqual','tolerance']

# distribution object
class Distribution(object):
    """
Sampling distribution for mystic optimizers
    """
    def __init__(self, generator=None, *args, **kwds):
        """
generate a sampling distribution with interface dist(size=None)

input::
    - generator: a 'distribution' method from scipy.stats or numpy.random
    - rng: a mystic.random_state object [default: random_state('numpy.random')]
    - args: positional arguments for the distribtution object
    - kwds: keyword arguments for the distribution object

note::
    this method only accepts numpy.random methods with the keyword 'size',
    and only accepts random_state objects built with module='numpy.random'
        """ #XXX: generate Distribution from list of Distributions?
        from mystic.tools import random_state
        rng = kwds.pop('rng', random_state(module='numpy.random'))
        if generator is None: generator = rng.random
        if getattr(generator, 'rvs', False): 
            d = generator(*args, **kwds)
            self.rvs = lambda size=None: d.rvs(size=size, random_state=rng)
        else:
            d = getattr(rng, generator.__name__)
            self.rvs = lambda size=None: d(size=size, *args, **kwds)
        return
    def __call__(self, size=None):
        """generate a sample of given size (tuple) from the distribution"""
        return self.rvs(size)

# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
 
""" NOTE:
due to pickling issues, cost function is provided w/o using a factory method.
also, a local import of 'polyeval' avoides a NameError.
"""

def chebyshev8cost(trial,M=61):
    """The costfunction for order-n Chebyshev fitting.
M evaluation points between [-1, 1], and two end points"""

    from mystic.models.poly import chebyshev8coeffs as target
    from mystic.math import polyeval

    result=0.0
    x=-1.0
    dx = 2.0 / (M-1)
    for i in range(M):
        px = polyeval(trial, x)
        if px<-1 or px>1:
            result += (1 - px) * (1 - px)
        x += dx

    px = polyeval(trial, 1.2) - polyeval(target, 1.2)
    if px<0: result += px*px

    px = polyeval(trial, -1.2) - polyeval(target, -1.2)
    if px<0: result += px*px

    return result

if __name__=='__main__':
    from mystic.models.poly import chebyshev8coeffs as target
    from mystic.math import poly1d
    print(poly1d(target))
    print("")
    print(chebyshev8cost(target))



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Simple utility functions for SV-classifications
"""

from numpy import multiply, asarray, dot, transpose, sum

def KernelMatrix(X, k=dot): #XXX: generalized dot/inner product?  K(X,Y)?
    "inner product of X with self, using k as elementwise product function"
    # the following is tensordot(X,X,axes=(-1,-1)), with dot --> k
    # 3-clause BSD (see: v1.7.2 http://docs.scipy.org/doc/numpy/license.html)
    X = asarray(X)
    Xs = X.shape
    ndX = len(X.shape)
    nX = 1
    axes = [ndX - 1]

    # Move the axes to sum over to the end of "a"
    # and to the front of "b" in inner(a,b)
    notin = [_ for _ in range(ndX) if _ not in axes]
    newaxes_a = notin + axes
    N2 = 1
    for axis in axes: N2 *= Xs[axis]
    newshape_a = (-1, N2)
    olda = [Xs[axis] for axis in notin]

    newaxes_b = axes + notin
    N2 = 1
    for axis in axes: N2 *= Xs[axis]
    newshape_b = (N2, -1)
    oldb = [Xs[axis] for axis in notin]

    at = X.transpose(newaxes_a).reshape(newshape_a)
    bt = X.transpose(newaxes_b).reshape(newshape_b)
    return k(at,bt).reshape(olda + oldb)


def WeightVector(alpha, X, y):
    ay = (alpha * y).flatten()
    aXy = transpose(ay * transpose(X))
    return sum(aXy, 0)


def SupportVectors(alpha, y=None, epsilon=0):
    """indices of nonzero alphas (at tolerance epsilon)

If labels y are provided, then group indices by label
    """
    import mystic.svr as svr
    sv = svr.SupportVectors(alpha,epsilon)
    if y is None:
        return sv
    else:
        class1 = set((y>0).nonzero()[1])
        class2 = set((y<0).nonzero()[1])
        sv1 = class1.intersection(sv)
        sv2 = class2.intersection(sv)
        return list(sv1), list(sv2)

def Bias(alpha, X, y, kernel=dot):
    """Compute classification bias. """
    sv1, sv2 = SupportVectors(alpha, y, epsilon=1e-6)
    pt1, pt2 = X[sv1[0],:], X[sv2[0],:]
    k1, k2 = kernel(X, pt1), kernel(X,pt2)
    return -0.5 * (sum(alpha*y*k1) + sum(alpha*y*k2))

# end of file



#!/usr/bin/env python
#
# Problem definition:
# Example in reference documentation for scipy.optimize.linprog.
# http://docs.scipy.org/doc/scipy-dev/reference/optimize.linprog-simplex.html
# 
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
    Minimize: f = -1*x[0] + 4*x[1]

    Subject to: -3*x[0] + 1*x[1] <= 6
                 1*x[0] + 2*x[1] <= 4
                            x[1] >= -3

    where:  -inf <= x[0] <= inf
"""

def objective(x):
  x0,x1 = x
  return -x0 + 4*x1

equations = """
-3*x0 + x1 - 6.0 <= 0.0
x0 + 2*x1 - 4.0 <= 0.0
"""
bounds = [(None, None),(-3.0, None)]

# with penalty='penalty' applied, solution is:
xs = [10.0, -3.0]
ys = -22.0
# alternately, if solving for the maximum, the solution is:
_xs = [-1.14285714,  2.57142857]
_ys = 11.428571428571429

from mystic.symbolic import generate_conditions, generate_penalty
pf = generate_penalty(generate_conditions(equations))
from mystic.symbolic import generate_constraint, generate_solvers, simplify
cf = generate_constraint(generate_solvers(simplify(equations)))

# inverted objective, used in solving for the maximum
_objective = lambda x: -objective(x)


if __name__ == '__main__':

  from mystic.solvers import diffev2, fmin_powell
  from mystic.math import almostEqual

  result = diffev2(objective, x0=bounds, bounds=bounds, constraint=cf, penalty=pf, npop=40, disp=False, full_output=True)
  assert almostEqual(result[0], xs, rel=1e-2)
  assert almostEqual(result[1], ys, rel=1e-2)

  result = fmin_powell(objective, x0=[0.0,0.0], bounds=bounds, constraint=cf, penalty=pf, disp=False, full_output=True)
  assert almostEqual(result[0], xs, rel=1e-2)
  assert almostEqual(result[1], ys, rel=1e-2)

  # alternately, solving for the maximum
  result = diffev2(_objective, x0=bounds, bounds=bounds, constraint=cf, penalty=pf, npop=40, disp=False, full_output=True)
  assert almostEqual( result[0], _xs, rel=1e-2)
  assert almostEqual(-result[1], _ys, rel=1e-2)

  result = fmin_powell(_objective, x0=[0,0], bounds=bounds, constraint=cf, penalty=pf, npop=40, disp=False, full_output=True)
  assert almostEqual( result[0], _xs, rel=1e-2)
  assert almostEqual(-result[1], _ys, rel=1e-2)


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2019-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
'''higher-level readers and writers for stored functions
'''

__all__ = ['db','write','read']

from . import archive as rb


def db(name):
    """get function db with the name 'name'"""
    if not isinstance(name, (str, (u'').__class__)):
        return getattr(name, '__archive__', name) # need cached == False
    return rb._read_func(name) #, type=type)


def write(function, archives):
    """write function to corresponding archives"""
    if isinstance(archives, (str, (u'').__class__)):
        archives = db(archives)
    if getattr(function, '__axis__', None) is None: #XXX: and no len(archives)
        rb._write_func(archives, function, {})
    else:
        [rb._write_func(a, f, {}) for a,f in zip(archives,function.__axis__)]


def read(archives): # method?
    """read stored functions from the list of dbs

    Args:
        archives (list[string]): list of names of function archives

    Returns:
        a klepto.archive instance

    Notes:
        The order of the dbs is important, with the index of ``archives`` 
        corresponding to the desired axis. If a db is empty, returns ``None``
        for the empty db. Also, a klepto.archive instance can be provided
        instead of the ``name`` of the db.
    """
    if isinstance(archives, (str, (u'').__class__)):
        archives = db(archives)
    if type(archives).__module__.startswith('klepto.'):
        f = rb.read_func(db(archives))
        if f is None:
            return f
        f = f[0]
        if not hasattr(f, '__axis__'):
            f.__axis__ = None
        return f
    from mystic.math.interpolate import interpf
    f = interpf([[1,2],[2,3]],[[1,2],[2,3]],method='thin_plate')
    f.__axis__[:] = [(i if i is None else i[0]) for i in map(rb.read_func, map(db, archives))]
    return f




#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
__doc__ = """
functional interfaces for mystic's visual diagnistics for support files
"""

__all__ = ['convergence', 'hypercube', 'hypercube_measures', \
           'hypercube_scenario', 'best_dimensions', 'swap']

# from mpl_toolkits.mplot3d import axes3d

# globals
__quit = False
ZERO = 1.0e-6  # zero-ish
import sys
if (sys.hexversion >= 0x30000f0):
    PY3 = True
    exec_string = 'exec(code, globals)'
else:
    PY3 = False
    exec_string = 'exec code in globals'
NL = '\n'
NL2 = '\n\n'
#FIXME: remove this head-standing to workaround python2.6 exec bug

def best_dimensions(n):
  """get the 'best' dimensions ``(i x j)`` for arranging plots

Args:
  n (int): number of plots

Returns:
  tuple ``(i,j)`` of ``i`` rows ``j`` columns, where ``i*j`` is roughly ``n``
  """
  from mystic.tools import factor
  allfactors = list(factor(n))
  from numpy import product
  cand = [1] + [product(allfactors[:i+1]) for i in range(len(allfactors))]
 #return cand[-1], n/cand[-1]
  best = [cand[len(cand)//2], n//cand[len(cand)//2]]
  best.sort(reverse=True)
  return tuple(best)
# if len(cand)%2:
#   return cand[len(cand)/2], cand[len(cand)/2]
# return cand[len(cand)/2], cand[len(cand)/2 - 1]


#### building the cone primitive ####
def _cone_builder(slope, bounds, strict=True):
  """factory to create a cone primitive

  slope -- slope multiplier for cone on the X,Y,Z axes (for mesh construction)
  bounds -- list of tuples of bounds for the plot; (lower,upper) for each axis
"""
  from mystic.math import almostEqual
  import numpy as np
  def cone_mesh(length):
    """construct a conical mesh for a given length of cone"""
    L1,L2,L3 = slope
    radius = length / L3 #XXX: * 0.5
    r0 = ZERO

    if almostEqual(radius, r0, tol=r0): radius = r0
    r = np.linspace(radius,radius,6) 
    r[0]= np.zeros(r[0].shape) 
    r[1] *= r0/radius
    r[5] *= r0/radius
    r[3]= np.zeros(r[3].shape) 

    p = np.linspace(0,2*np.pi,50) 
    R,P = np.meshgrid(r,p) 
    X,Y = L1 * R*np.cos(P), L2 * R*np.sin(P) 

    tmp=list() 
    for i in range(np.size(p)): 
      tmp.append([0,0,length,length,length,0]) # len = size(r)
    Z = np.array(tmp) 
    return X,Z,Y

  lb,ub = zip(*bounds)
  # if False, the cone surface may violate bounds
 #strict = True # always respect the bounds

  def cone(position, top=True):
    """ construct a cone primitive, with vertex at given position

    position -- tuple of vertex position
    top -- boolean for if 'top' or 'bottom' cone
"""
    from numpy import asarray
    position = asarray(position)
    d_hi = ub - position
    d_lo = lb - position
    if strict and (any(d_hi < 0) or any(d_lo > 0)):
      return None  # don't plot cone if vertex is out of bounds

    if top: # distance of vertex from upper edge
      l = d_hi[2]
      if not l: l = ZERO
    else:   # distance of vertex from lower edge
      l = d_lo[2]
      if not l: l = -ZERO
    X,Z,Y = cone_mesh(l)
    X = X + position[0]
    Y = Y + position[1]
    Z = Z + position[2]
    return X,Z,Y

  return cone


#### plotting the cones ####
def _plot_bowtie(ax, data, slope, bounds, color='0.75', axis=None, tol=0.0):
  """plot (2D) double cones for a given dataset

  ax -- matplotlib 'Axes3D' plot object
  data -- list of datapoints, where datapoints are 3-tuples (i.e. x,y,z)
  slope -- slope multiplier for cone on the X,Y,Z axes (for mesh construction)
  bounds -- list of tuples of bounds for the plot; (lower,upper) for each axis
  color -- string name (or rbg value) of color to use for datapoints
  axis -- the axis of the cone
  tol -- distance between center of mass of the double cones and a cone vertex
"""
  if axis not in range(len(bounds)-1): return ax
  from numpy import asarray, inf
  data = asarray(data)
  sl = slope[axis]
  gap = max(0., tol)
  # find the endpoints of the cone:  y = slope * (x0 - x) + y0
  ylo = sl * (bounds[0][0] - data.T[0]) + data.T[1]
  yhi = sl * (bounds[0][1] - data.T[0]) + data.T[1]
  zhi = -sl * (bounds[0][0] - data.T[0]) + data.T[1]
  zlo = -sl * (bounds[0][1] - data.T[0]) + data.T[1]
  xlb = bounds[0][0]
  xub = bounds[0][1]
  ylb = bounds[1][0]
  yub = bounds[1][1]

  for pt,yl,yu,zl,zu in zip(data,ylo,yhi,zlo,zhi):
   #ax.plot([xlb, pt[0], xub], [yl, pt[1], yu], color='black')
   #ax.plot([xlb, pt[0], xub], [zu, pt[1], zl], color='black')
    ax.fill_between([xlb, pt[0], xub], [ylb]*3, [yl-gap, pt[1]-gap, zl-gap], \
                                     facecolor=color, alpha=0.2)
    ax.fill_between([xlb, pt[0], xub], [yub]*3, [zu+gap, pt[1]+gap, yu+gap], \
                                     facecolor=color, alpha=0.2)
  return ax


def _plot_cones(ax, data, slope, bounds, color='0.75', axis=None, strict=True, tol=0.0):
  """plot (3D) double cones for a given dataset

  ax -- matplotlib 'Axes3D' plot object
  data -- list of datapoints, where datapoints are 3-tuples (i.e. x,y,z)
  slope -- slope multiplier for cone on the X,Y,Z axes (for mesh construction)
  bounds -- list of tuples of bounds for the plot; (lower,upper) for each axis
  color -- string name (or rbg value) of color to use for datapoints
  axis -- the axis of the cone
  tol -- distance between center of mass of the double cones and a cone vertex
"""
  from numpy import asarray, zeros
  # adjust slope, bounds, and data so cone axis is last 
  slope = swap(slope, axis) 
  bounds = swap(bounds, axis) 
  data = [swap(pt,axis) for pt in data]
  cone = _cone_builder(slope, bounds, strict=strict)
  # prepare to apply the 'gap' tolerance
  gap = zeros(len(slope))
  gap[-1:] = [max(0., tol)] #XXX: bad behavior if len(slope) is 0
  # plot the upper and lower cone
  for datapt in data:
    datapt = asarray(datapt)
    _cone = cone(datapt + gap, top=True)
    if _cone:
      X,Z,Y = swap(_cone, axis) # 'unswap' the axes
      ax.plot_surface(X, Y,Z, rstride=1, cstride=1, color=color, \
                                         linewidths=0, alpha=.1)
    _cone = cone(datapt - gap, top=False)
    if _cone:
      X,Z,Y = swap(_cone, axis) # 'unswap' the axes
      ax.plot_surface(X, Y,Z, rstride=1, cstride=1, color=color, \
                                         linewidths=0, alpha=.1)
  return ax


def _plot_data(ax, data, bounds, color='red', strict=True, **kwds):
  """plot datapoints for a given dataset

  ax -- matplotlib 'Axes3D' plot object
  data -- list of datapoints, where datapoints are 3-tuples (i.e. x,y,z)
  bounds -- list of tuples of bounds for the plot; (lower,upper) for each axis
  color -- string name (or rbg value) of color to use for datapoints
"""
  _2D = kwds.get('_2D', False)
# strict = True # always respect the bounds
  lb,ub = zip(*bounds)
  # plot the datapoints themselves
  from numpy import asarray
  for datapt in data:
    position = asarray(datapt)
    d_hi = ub - position
    d_lo = lb - position
    if strict and (any(d_hi < 0) or any(d_lo > 0)): #XXX: any or just in Y ?
      pass  # don't plot if datapt is out of bounds
    else:
      if _2D:
        ax.plot([datapt[0]], [datapt[1]], \
                color=color, marker='o', markersize=10)
      else:
        ax.plot([datapt[0]], [datapt[1]], [datapt[2]], \
                color=color, marker='o', markersize=10)
  return ax


def _clip_axes(ax, bounds):
  """ clip plots to be within given lower and upper bounds

  ax -- matplotlib 'Axes3D' plot object
  bounds -- list of tuples of bounds for the plot; (lower,upper) for each axis
"""
  lb,ub = zip(*bounds)
  # plot only within [lb,ub]
  ax.set_xlim3d(lb[0], ub[0])
  ax.set_ylim3d(lb[1], ub[1])
  ax.set_zlim3d(lb[2], ub[2]) # cone "center" axis
  return ax


def _label_axes(ax, labels, **kwds):
  """ label plots with given string labels

  ax -- matplotlib 'Axes3D' plot object
  labels -- list of string labels for the plot
"""
  _2D = kwds.get('_2D', False)
  ax.set_xlabel(labels[0])
  ax.set_ylabel(labels[1])
  if not _2D:
    ax.set_zlabel(labels[2]) # cone "center" axis
  return ax


def _get_slope(data, replace=None, mask=None):
  """ replace one slope in a list of slopes with '1.0'
  (i.e. replace a slope from the dataset with the unit slope)

  data -- dataset object, where coordinates are 3-tuples and values are floats
  replace -- selected axis (an int) to plot values NOT coords
  """
  slope = data.lipschitz
  if mask in range(len(slope)):
    slope = swap(slope, mask)
  if replace not in range(len(slope)):  # don't replace an axis
    return slope
  return slope[:replace] + [1.0] + slope[replace+1:]


def _get_coords(data, replace=None, mask=None):
  """ replace one coordiate axis in a 3-D data set with 'data values'
  (i.e. replace an 'x' axis with the 'y' values of the data)

  data -- dataset object, where coordinates are 3-tuples and values are floats
  replace -- selected axis (an int) to plot values NOT coords
  """
  slope = data.lipschitz
  coords = data.coords
  values = data.values
  if mask in range(len(slope)):
    coords = [swap(pt,mask) for pt in coords]
  if replace not in range(len(slope)):  # don't replace an axis
    return coords
  return [list(coords[i][:replace]) + [values[i]] + \
          list(coords[i][replace+1:]) for i in range(len(coords))]


def swap(alist, index=None):
  """swap the selected list element with the last element in alist

Args:
  alist (list): a list of objects
  index (int, default=None): the selected element

Returns:
  list with the elements swapped as indicated
  """
  if index not in range(len(alist)):  # don't swap an element
    return alist 
  return alist[:index] + alist[index+1:] + alist[index:index+1]


def_convergence = '''
def convergence(filename, **kwds):
    """
generate parameter convergence plots from file written with ``write_support_file``

Available from the command shell as::

    support_convergence filename [options]

or as a function call::

    mystic.support.convergence(filename, **options)

Args:
    filename (str): name of the convergence logfile (e.g ``paramlog.py``)

Returns:
    None

Notes:
    - The option *out* takes a string of the filepath for the generated plot.
      If ``out = True``, return the Figure object instead of generating a plot.
    - The option *iter* takes an integer of the largest iteration to plot.
    - The option *param* takes an indicator string. The indicator string is
      built from comma-separated array slices. For example, ``param = ":"``
      will plot all parameters in a single plot.  Alternatively,
      ``param = ":2, 2:"`` will split the parameters into two plots, and
      ``param = "0"`` will only plot the first parameter.
    - The option *label* takes comma-separated strings. For example,
      ``label = "x,y,"`` will label the y-axis of the first plot with 'x', a
      second plot with 'y', and not add a label to a third or subsequent plots.      If more labels are given than plots, then the last label will be used
      for the y-axis of the 'cost' plot. LaTeX is also accepted. For example,
      ``label = "$ h$, $ a$, $ v$"`` will label the axes with standard
      LaTeX math formatting. Note that the leading space is required, and the
      text is aligned along the axis.
    - The option *nid* takes an integer of the nth simultaneous points to plot.
    - The option *cost* takes a boolean, and will also plot the parameter cost.
    - The option *legend* takes a boolean, and will display the legend.
"""
    import shlex
    try:
        basestring
        from StringIO import StringIO
    except NameError:
        basestring = str
        from io import StringIO
    global __quit
    __quit = False
    _out = False

    instance = None
    # handle the special case where list is provided by sys.argv
    if isinstance(filename, (list,tuple)) and not kwds:
        cmdargs = filename # (above is used by script to parse command line)
    elif isinstance(filename, basestring) and not kwds:
        cmdargs = shlex.split(filename)
    # 'everything else' is essentially the functional interface
    else:
        cmdargs = kwds.get('kwds', '')
        if not cmdargs:
            out = kwds.get('out', None)
            iter = kwds.get('iter', None)
            param = kwds.get('param', None)
            label = kwds.get('label', None)
            nid = kwds.get('nid', None)
            cost = kwds.get('cost', False)
            legend = kwds.get('legend', False)

            if isinstance(out, bool): _out, out = out, None

            # process "commandline" arguments
            cmdargs = ''
            cmdargs += '' if out is None else '--out={} '.format(out)
            cmdargs += '' if iter is None else '--iter={} '.format(iter)
            cmdargs += '' if param is None else '--param="{}" '.format(param)
            cmdargs += '' if label is None else '--label="{}" '.format(label)
            cmdargs += '' if nid is None else '--nid={} '.format(nid)
            cmdargs += '' if cost == False else '--cost '
            cmdargs += '' if legend == False else '--legend '
        else:
            cmdargs = ' ' + cmdargs
        if isinstance(filename, basestring):
            cmdargs = filename.split() + shlex.split(cmdargs)
        else: # special case of passing in monitor instance
            instance = filename
            cmdargs = shlex.split(cmdargs)

    #XXX: note that 'argparse' is new as of python2.7
    from optparse import OptionParser
    def _exit(self, errno=None, msg=None):
        global __quit
        __quit = True
        if errno or msg:
            msg = msg.split(': error: ')[-1].strip()
            raise IOError(msg)
    OptionParser.exit = _exit

    parser = OptionParser(usage=convergence.__doc__.split(NL2+'Options:')[0])
    parser.add_option("-u","--out",action="store",dest="out",\
                      metavar="STR",default=None,
                      help="filepath to save generated plot")
    parser.add_option("-i","--iter",action="store",dest="step",metavar="INT",\
                      default=None,help="the largest iteration to plot")
    parser.add_option("-p","--param",action="store",dest="param",\
                      metavar="STR",default=":",
                      help="indicator string to select parameters")
    parser.add_option("-l","--label",action="store",dest="label",\
                      metavar="STR",default="",
                      help="string to assign label to y-axis")
    parser.add_option("-n","--nid",action="store",dest="id",\
                      metavar="INT",default=None,
                      help="id # of the nth simultaneous points to plot")
    parser.add_option("-c","--cost",action="store_true",dest="cost",\
                      default=False,help="also plot the parameter cost")
    parser.add_option("-g","--legend",action="store_true",dest="legend",\
                      default=False,help="show the legend")

    if PY3:
        f = StringIO()
        parser.print_help(file=f)
        f.seek(0)
        if 'Options:' not in convergence.__doc__:
            convergence.__doc__ += NL+'Options:{0}'.format(f.read().split('Options:')[-1])
        f.close()
    else:
        if 'Options:' not in convergence.__doc__:
            convergence.__doc__ += NL+'Options:{0}'.format(parser.format_help().split('Options:')[-1])

    try:
        parsed_opts, parsed_args = parser.parse_args(cmdargs)
    except UnboundLocalError:
        pass
    if __quit: return

    # get the name of the parameter log file
    if instance is None:
        instance = parsed_args[0]
    from mystic.munge import read_history
    params, cost = read_history(instance)

    if parsed_opts.cost: # also plot the cost
       #exec "from {file} import cost".format(file=file)
        pass
    else:
        cost = None

    try: # path of plot output file
      out = parsed_opts.out  # e.g. 'myplot.png'
      if "None" == out: out = None
    except:
      out = None

    if parsed_opts.legend: # show the legend
        legend = True
    else:
        legend = False

    try: # select which iteration to stop plotting at
        step = int(parsed_opts.step)
    except:
        step = None

    try: # select which parameters to plot
        select = parsed_opts.param.split(',')  # format is ":2, 2:4, 5, 6:"
    except:
        select = [':']
       #select = [':1']
       #select = [':2','2:']
       #select = [':1','1:2','2:3','3:']
       #select = ['0','1','2','3']
    plots = len(select)

    try: # select labels for the axes
        label = parsed_opts.label.split(',')  # format is "x, y, z"
        label += [''] * max(0, plots - len(label))
    except:
        label = [''] * plots

    try: # select which 'id' to plot results for
        id = int(parsed_opts.id)
    except:
        id = None # i.e. 'all' **or** use id=0, which should be 'best' energy ?

    # ensure all terms of select have a ":"
    for i in range(plots):
        if isinstance(select[i], int): select[i] = str(select[i])
        if select[i] == '-1': select[i] = 'len(params)-1:len(params)'
        elif not select[i].count(':'):
            select[i] += ':' + str(int(select[i])+1)

    # take only the first 'step' iterations
    params = [var[:step] for var in params]
    if cost:
        cost = cost[:step]

    # take only the selected 'id'
    if id != None:
        param = []
        for j in range(len(params)):
            param.append([p[id] for p in params[j]])
        params = param[:]

    # handle special case where only plot the cost
    if cost and len(select) == 1 and select[0].endswith(':0'):
        dim1,dim2 = best_dimensions(1)

        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax1 = fig.add_subplot(dim1,dim2,1)
        ax1.set_ylabel(label[-1])
        ax1.plot(cost,label='cost')
        if legend: plt.legend()
        cost = None
    else:
        if cost: j = 1
        else: j = 0
        dim1,dim2 = best_dimensions(plots + j)

        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax1 = fig.add_subplot(dim1,dim2,1)
        ax1.set_ylabel(label[0])
        locals = dict(params=params)
        data = eval("params[{0}]".format(select[0]), locals)
        try:
            n = int(select[0].split(":")[0])
        except ValueError:
            n = 0
        for line in data:
            ax1.plot(line,label=str(n))#, marker='o')
            n += 1
        if legend: plt.legend()

    globals = {'fig':fig, 'dim1':dim1, 'dim2':dim2, 'ax1':ax1, 'label':label}
    for i in range(2, plots + 1):
        code = "ax{0:d} = fig.add_subplot(dim1,dim2,{0:d}, sharex=ax1);".format(i)
        code += "ax{0:d}.set_ylabel(label[{1:d}]);".format(i,i-1)
        code = compile(code, '<string>', 'exec')
        %(exec_string)s
        data = eval("params[{0}]".format(select[i-1]), locals)
        try:
            n = int(select[i-1].split(":")[0])
        except ValueError:
            n = 0
        for line in data:
            globals['line'] = line
            code = "ax{0:d}.plot(line,label='{1}')#, marker='o')".format(i,n)
            code = compile(code, '<string>', 'exec')
            %(exec_string)s
            n += 1
        if legend: plt.legend()
    if cost:
        globals['cost'] = cost
        code = "cx1 = fig.add_subplot(dim1,dim2,{0:d}, sharex=ax1);".format(int(plots+1))
        code += "cx1.plot(cost,label='cost');"#, marker='o')"
        if max(0, len(label) - plots): code += "cx1.set_ylabel(label[-1]);"
        code = compile(code, '<string>', 'exec')
        %(exec_string)s
        if legend: plt.legend()

    # process inputs
    if _out: out = _out
    if out: out = _out or out

    if not out:
        plt.show()
    elif out is True:
        return fig #XXX: better?: fig.axes if len(fig.axes) > 1 else ax
    else:
        fig.savefig(out)
''' % dict(exec_string=exec_string)

    ### USUAL WAY OF CREATING PLOTS ###
    #fig = plt.figure()
    #ax1 = fig.add_subplot(3,2,1)
    ##ax1.ylim(60,105)
    #ax1.plot(x)
    #ax1.plot(x2)
    #plt.title('convergence for thickness support')
    ##plt.xlabel('iterations')
    #plt.ylabel('thickness')
    #
    #ax2 = fig.add_subplot(3,2,2, sharex=ax1)
    ##ax2.ylim(0,1)
    #ax2.plot(wx)
    #ax2.plot(wx2)
    #plt.title('convergence for weight(thickness)')
    ##plt.xlabel('iterations')
    #plt.ylabel('weight')
    #
    #plt.show()
    ###################################


def_hypercube = '''
def hypercube(filename, **kwds):
    """
generate parameter support plots from file written with ``write_support_file``

Available from the command shell as::

    support_hypercube filename [options]

or as a function call::

    mystic.support.hypercube(filename, **options)

Args:
    filename (str): name of the convergence logfile (e.g ``paramlog.py``)

Returns:
    None

Notes:
    - The option *out* takes a string of the filepath for the generated plot.
      If ``out = True``, return the Figure object instead of generating a plot.
    - The options *bounds*, *axes*, and *iters* all take indicator strings.
      The bounds should be given as comma-separated slices. For example, using
      ``bounds = "60:105, 0:30, 2.1:2.8"`` will set the lower and upper bounds
      for x to be (60,105), y to be (0,30), and z to be (2.1,2.8). Similarly,
      axes also accepts comma-separated groups of ints; however, for axes, each      entry indicates which parameters are to be plotted along each axis -- the
      first group for the x direction, the second for the y direction, and
      third for z. Thus, ``axes = "2 3, 6 7, 10 11"`` would set 2nd and 3rd
      parameters along x. Iters also accepts strings built from comma-separated      array slices. For example, ``iters = ":"`` will plot all iters in a
      single plot. Alternatively, ``iters = ":2, 2:"`` will split the iters
      into two plots, while ``iters = "0"`` will only plot the first iteration.
    - The option *label* takes comma-separated strings. Thus ``label = "x,y,"``      will place 'x' on the x-axis, 'y' on the y-axis, and nothing on the
      z-axis. LaTeX, such as ``label = "$ h $, $ a$, $ v$"`` will label the
      axes with standard LaTeX math formatting. Note that the leading space is
      required, while a trailing space aligns the text with the axis instead of      the plot frame.
    - The option *nid* takes an integer of the nth simultaneous points to plot.
    - The option *scale* takes an integer as a grayscale contrast multiplier.
    - The option *flat* takes a boolean, to plot results in a single plot.
"""
    import shlex
    try:
        basestring
        from StringIO import StringIO
    except NameError:
        basestring = str
        from io import StringIO
    global __quit
    __quit = False
    _out = False

    instance = None
    # handle the special case where list is provided by sys.argv
    if isinstance(filename, (list,tuple)) and not kwds:
        cmdargs = filename # (above is used by script to parse command line)
    elif isinstance(filename, basestring) and not kwds:
        cmdargs = shlex.split(filename)
    # 'everything else' is essentially the functional interface
    else:
        cmdargs = kwds.get('kwds', '')
        if not cmdargs:
            out = kwds.get('out', None)
            bounds = kwds.get('bounds', None)
            axes = kwds.get('axes', None)
            iters = kwds.get('iters', None)
            label = kwds.get('label', None)
            nid = kwds.get('nid', None)
            scale = kwds.get('scale', None)
            flat = kwds.get('flat', False)

            if isinstance(out, bool): _out, out = out, None

            # process "commandline" arguments
            cmdargs = ''
            cmdargs += '' if out is None else '--out={} '.format(out)
            cmdargs += '' if bounds is None else '--bounds="{}" '.format(bounds)
            cmdargs += '' if axes is None else '--axes="{}" '.format(axes)
            cmdargs += '' if iters is None else '--iters="{}" '.format(iters)
            cmdargs += '' if label is None else '--label="{}" '.format(label)
            cmdargs += '' if nid is None else '--nid={} '.format(nid)
            cmdargs += '' if scale is None else '--scale={} '.format(scale)
            cmdargs += '' if flat == False else '--flat '
        else:
            cmdargs = ' ' + cmdargs
        if isinstance(filename, basestring):
            cmdargs = filename.split() + shlex.split(cmdargs)
        else: # special case of passing in monitor instance
            instance = filename
            cmdargs = shlex.split(cmdargs)

    #XXX: note that 'argparse' is new as of python2.7
    from optparse import OptionParser
    def _exit(self, errno=None, msg=None):
        global __quit
        __quit = True
        if errno or msg:
            msg = msg.split(': error: ')[-1].strip()
            raise IOError(msg)
    OptionParser.exit = _exit

    parser = OptionParser(usage=hypercube.__doc__.split(NL2+'Options:')[0])
    parser.add_option("-u","--out",action="store",dest="out",\
                      metavar="STR",default=None,
                      help="filepath to save generated plot")
    parser.add_option("-b","--bounds",action="store",dest="bounds",\
                      metavar="STR",default="0:1, 0:1, 0:1",
                      help="indicator string to set hypercube bounds")
    parser.add_option("-x","--axes",action="store",dest="xyz",\
                      metavar="STR",default="0, 1, 2",
                      help="indicator string to assign parameter to axis")
    parser.add_option("-i","--iters",action="store",dest="iters",\
                      metavar="STR",default="-1",
                      help="indicator string to select iterations to plot")
    parser.add_option("-l","--label",action="store",dest="label",\
                      metavar="STR",default=",,",
                      help="string to assign label to axis")
    parser.add_option("-n","--nid",action="store",dest="id",\
                      metavar="INT",default=None,
                      help="id # of the nth simultaneous points to plot")
    parser.add_option("-s","--scale",action="store",dest="scale",\
                      metavar="INT",default=1.0,
                      help="grayscale contrast multiplier for points in plot")
    parser.add_option("-f","--flat",action="store_true",dest="flatten",\
                      default=False,help="show selected iterations in a single plot")

    if PY3:
        f = StringIO()
        parser.print_help(file=f)
        f.seek(0)
        if 'Options:' not in hypercube.__doc__:
            hypercube.__doc__ += NL+'Options:{0}'.format(f.read().split('Options:')[-1])
        f.close()
    else:
        if 'Options:' not in hypercube.__doc__:
            hypercube.__doc__ += NL+'Options:{0}'.format(parser.format_help().split('Options:')[-1])

    try:
        parsed_opts, parsed_args = parser.parse_args(cmdargs)
    except UnboundLocalError:
        pass
    if __quit: return

    # get the name of the parameter log file
    if instance is None:
        instance = parsed_args[0]
    from mystic.munge import read_history
    params, _cost = read_history(instance)
    # would be nice to use meta = ['wx','wx2','x','x2','wy',...]
    # exec "from {file} import meta".format(file=file)

    try: # select the bounds
        bounds = parsed_opts.bounds.split(",")  # format is "60:105, 0:30, 2.1:2.8"
        bounds = [tuple(float(j) for j in i.split(':')) for i in bounds]
    except:
        bounds = [(0,1),(0,1),(0,1)]

    try: # select which params are along which axes
        xyz = parsed_opts.xyz.split(",")  # format is "0 1, 4 5, 8 9"
        xyz = [tuple(int(j) for j in i.split()) for i in xyz]
    except:
        xyz = [(0,),(1,),(2,)]

    try: # select labels for the axes
        label = parsed_opts.label.split(',')  # format is "x, y, z"
    except:
        label = ['','','']

    x = params[max(xyz[0])]
    try: # select which iterations to plot
        select = parsed_opts.iters.split(',')  # format is ":2, 2:4, 5, 6:"
    except:
        select = ['-1']
       #select = [':']
       #select = [':1']
       #select = [':2','2:']
       #select = [':1','1:2','2:3','3:']
       #select = ['0','1','2','3']

    try: # path of plot output file
      out = parsed_opts.out  # e.g. 'myplot.png'
      if "None" == out: out = None
    except:
      out = None

    try: # collapse non-consecutive iterations into a single plot...
        flatten = parsed_opts.flatten
    except:
        flatten = False

    try: # select which 'id' to plot results for
        id = int(parsed_opts.id)
    except:
        id = None # i.e. 'all' **or** use id=0, which should be 'best' energy ?

    try: # scale the color in plotting the weights
        scale = float(parsed_opts.scale)
    except:
        scale = 1.0 # color = color**scale

    # ensure all terms of bounds and xyz are tuples
    for bound in bounds:
        if not isinstance(bound, tuple):
            raise TypeError("bounds should be tuples of (lower_bound,upper_bound)")
    for i in range(len(xyz)):
        if isinstance(xyz[i], int):
            xyz[i] = (xyz[i],)
        elif not isinstance(xyz[i], tuple):
            raise TypeError("xyz should be tuples of (param1,param2,param3,...)")

    # ensure all terms of select are strings that have a ":"
    for i in range(len(select)):
        if isinstance(select[i], int): select[i] = str(select[i])
        if select[i] == '-1': select[i] = 'len(x)-1:len(x)'
        elif not select[i].count(':'):
            select[i] += ':' + str(int(select[i])+1)

    # take only the selected 'id'
    if id != None:
        param = []
        for j in range(len(params)):
            param.append([p[id] for p in params[j]])
        params = param[:]

    # at this point, we should have:
    #bounds = [(60,105),(0,30),(2.1,2.8)] or [(None,None)]*3
    #xyz = [(0,1),(4,5),(8,9)] for any length tuple
    #select = ['-1:'] or [':'] or [':1','1:2','2:3','3:'] or similar
    #id = 0 or None

    plots = len(select)
    if not flatten:
        dim1,dim2 = best_dimensions(plots)
    else: dim1,dim2 = 1,1

    # use the default bounds where not specified
    bounds = [list(i) for i in bounds]
    for i in range(len(bounds)):
        if bounds[i][0] is None: bounds[i][0] = 0
        if bounds[i][1] is None: bounds[i][1] = 1

    # correctly bound the first plot.  there must be at least one plot
    import matplotlib.pyplot as plt
    fig = plt.figure()
    try:
        Subplot3D = None
        ax1 = fig.add_subplot(dim1, dim2, 1, projection='3d')
        newsubplotcode = "ax = ax{0:d} = fig.add_subplot(dim1,dim2,{0:d}, projection='3d');"
    except:
        from mpl_toolkits.mplot3d import Axes3D as _Axes3D
        from matplotlib.axes import subplot_class_factory
        Subplot3D = subplot_class_factory(_Axes3D)
        ax1 = Subplot3D(fig, dim1,dim2,1)
        newsubplotcode = "ax = ax{0:d} = Subplot3D(fig, dim1,dim2,{0:d});"
    ax1.plot([bounds[0][0]],[bounds[1][0]],[bounds[2][0]])
    ax1.plot([bounds[0][1]],[bounds[1][1]],[bounds[2][1]])
    globals = {'plt':plt,'fig':fig,'dim1':dim1,'dim2':dim2,\
               'Subplot3D':Subplot3D,'bounds':bounds,'label':label}
    if not flatten:
        code = "plt.title('iterations[{0}]');".format(select[0])
    else: 
        code = "plt.title('iterations[*]');"
    code = compile(code, '<string>', 'exec')
    %(exec_string)s
    ax1.set_xlabel(label[0])
    ax1.set_ylabel(label[1])
    ax1.set_zlabel(label[2])
    a = [ax1]

    # set up additional plots
    if not flatten:
        for i in range(2, plots + 1):
            code = newsubplotcode.format(i)
            code += "ax.plot([bounds[0][0]],[bounds[1][0]],[bounds[2][0]]);"
            code += "ax.plot([bounds[0][1]],[bounds[1][1]],[bounds[2][1]]);"
            code += "plt.title('iterations[{0}]');".format(select[i - 1])
            code += "ax.set_xlabel(label[0]);"
            code += "ax.set_ylabel(label[1]);"
            code += "ax.set_zlabel(label[2]);"
            code = compile(code, '<string>', 'exec')
            %(exec_string)s
            a.append(globals['ax'])

    # turn each "n:m" in select to a list
    _select = []
    for sel in select:
        if sel[0] == ':': _select.append("0"+sel)
        else: _select.append(sel)
    for i in range(len(_select)):
        if _select[i][-1] == ':': select[i] = _select[i]+str(len(x))
        else: select[i] = _select[i]
    for i in range(len(select)):
        p = select[i].split(":")
        if p[0][0] == '-': p[0] = "len(x)"+p[0]
        if p[1][0] == '-': p[1] = "len(x)"+p[1]
        select[i] = p[0]+":"+p[1]
    locals = dict(x=x, params=params, xyz=xyz)
    steps = [eval("list(range({0}))".format(sel.replace(":",",")), locals) for sel in select]

    # at this point, we should have:
    #xyz = [(0,1),(4,5),(8,9)] for any length tuple
    #steps = [[0,1],[1,2],[2,3],[3,4,5,6,7,8]] or similar
    if flatten:
        from mystic.tools import flatten
        steps = [list(flatten(steps))]

    # build all the plots
    from numpy import inf, e
    scale = e**(scale - 1.0)
    for v in range(len(steps)):
        if len(steps[v]) > 1: qp = float(max(steps[v]))
        else: qp = inf 
        for s in steps[v]:
            # dot color determined by number of simultaneous iterations
            t = str((s/qp)**scale)
            for i in eval("[params[q][{0}] for q in xyz[0]]".format(s), locals):
                for j in eval("[params[q][{0}] for q in xyz[1]]".format(s), locals):
                    for k in eval("[params[q][{0}] for q in xyz[2]]".format(s), locals):
                        a[v].plot(i,j,k,marker='o',color=t,ms=10)

    # process inputs
    if _out: out = _out
    if out: out = _out or out

    if not out:
        plt.show()
    elif out is True:
        return fig #XXX: better?: fig.axes if len(fig.axes) > 1 else ax
    else:
        fig.savefig(out)
''' % dict(exec_string=exec_string)


def_hypercube_measures = '''
def hypercube_measures(filename, **kwds):
    """
generate measure support plots from file written with ``write_support_file``

Available from the command shell as::

    support_hypercube_measures filename [options]

or as a function call::

    mystic.support.hypercube_measures(filename, **options)

Args:
    filename (str): name of the convergence logfile (e.g ``paramlog.py``)

Returns:
    None

Notes:
    - The option *out* takes a string of the filepath for the generated plot.
      If ``out = True``, return the Figure object instead of generating a plot.
    - The options *bounds*, *axes*, *weight*, and *iters* all take indicator
      strings. The bounds should be given as comma-separated slices. For
      example, using ``bounds = "60:105, 0:30, 2.1:2.8"`` will set lower and
      upper bounds for x to be (60,105), y to be (0,30), and z to be (2.1,2.8).      Similarly, axes also accepts comma-separated groups of ints; however, for      axes, each entry indicates which parameters are to be plotted along each
      axis -- the first group for the x direction, the second for the y
      direction, and third for z. Thus, ``axes = "2 3, 6 7, 10 11"`` would set
      2nd and 3rd parameters along x. The corresponding weights are used to
      color the measure points, where 1.0 is black and 0.0 is white. For
      example, using ``weight = "0 1, 4 5, 8 9"`` would use the 0th and 1st
      parameters to weight x. Iters is also similar, however only accepts
      comma-separated ints. Hence, ``iters = "-1"`` will plot the last
      iteration, while ``iters = "0, 300, 700"`` will plot the 0th, 300th, and
      700th in three plots.
    - The option *label* takes comma-separated strings. Thus ``label = "x,y,"``      will place 'x' on the x-axis, 'y' on the y-axis, and nothing on the
      z-axis. LaTeX, such as ``label = "$ h $, $ a$, $ v$"`` will label the
      axes with standard LaTeX math formatting. Note that the leading space is
      required, while a trailing space aligns the text with the axis instead of      the plot frame.
    - The option *nid* takes an integer of the nth simultaneous points to plot.
    - The option *scale* takes an integer as a grayscale contrast multiplier.
    - The option *flat* takes a boolean, to plot results in a single plot.

Warning:
    This function is intended to visualize weighted measures (i.e. weights
    and positions), where the weights must be normalized (to 1) or an error
    will be thrown.
"""
    import shlex
    try:
        basestring
        from StringIO import StringIO
    except NameError:
        basestring = str
        from io import StringIO
    global __quit
    __quit = False
    _out = False

    instance = None
    # handle the special case where list is provided by sys.argv
    if isinstance(filename, (list,tuple)) and not kwds:
        cmdargs = filename # (above is used by script to parse command line)
    elif isinstance(filename, basestring) and not kwds:
        cmdargs = shlex.split(filename)
    # 'everything else' is essentially the functional interface
    else:
        cmdargs = kwds.get('kwds', '')
        if not cmdargs:
            out = kwds.get('out', None)
            bounds = kwds.get('bounds', None)
            axes = kwds.get('axes', None)
            weight = kwds.get('weight', None)
            iters = kwds.get('iters', None)
            label = kwds.get('label', None)
            nid = kwds.get('nid', None)
            scale = kwds.get('scale', None)
            flat = kwds.get('flat', False)

            if isinstance(out, bool): _out, out = out, None

            # process "commandline" arguments
            cmdargs = ''
            cmdargs += '' if out is None else '--out={} '.format(out)
            cmdargs += '' if bounds is None else '--bounds="{}" '.format(bounds)
            cmdargs += '' if axes is None else '--axes="{}" '.format(axes)
            cmdargs += '' if weight is None else '--weight="{}" '.format(weight)
            cmdargs += '' if iters is None else '--iters="{}" '.format(iters)
            cmdargs += '' if label is None else '--label="{}" '.format(label)
            cmdargs += '' if nid is None else '--nid={} '.format(nid)
            cmdargs += '' if scale is None else '--scale={} '.format(scale)
            cmdargs += '' if flat == False else '--flat '
        else:
            cmdargs = ' ' + cmdargs
        if isinstance(filename, basestring):
            cmdargs = filename.split() + shlex.split(cmdargs)
        else: # special case of passing in monitor instance
            instance = filename
            cmdargs = shlex.split(cmdargs)

    #XXX: note that 'argparse' is new as of python2.7
    from optparse import OptionParser
    def _exit(self, errno=None, msg=None):
        global __quit
        __quit = True
        if errno or msg:
            msg = msg.split(': error: ')[-1].strip()
            raise IOError(msg)
    OptionParser.exit = _exit

    parser = OptionParser(usage=hypercube_measures.__doc__.split(NL2+'Options:')[0])
    parser.add_option("-u","--out",action="store",dest="out",\
                      metavar="STR",default=None,
                      help="filepath to save generated plot")
    parser.add_option("-b","--bounds",action="store",dest="bounds",\
                      metavar="STR",default="0:1, 0:1, 0:1",
                      help="indicator string to set hypercube bounds")
    parser.add_option("-x","--axes",action="store",dest="xyz",\
                      metavar="STR",default="1, 3, 5",
                      help="indicator string to assign spatial parameter to axis")
    parser.add_option("-w","--weight",action="store",dest="wxyz",\
                      metavar="STR",default="0, 2, 4",
                      help="indicator string to assign weight parameter to axis")
    parser.add_option("-i","--iters",action="store",dest="iters",\
                      metavar="STR",default="-1",
                      help="indicator string to select iterations to plot")
    parser.add_option("-l","--label",action="store",dest="label",\
                      metavar="STR",default=",,",
                      help="string to assign label to axis")
    parser.add_option("-n","--nid",action="store",dest="id",\
                      metavar="INT",default=None,
                      help="id # of the nth simultaneous points to plot")
    parser.add_option("-s","--scale",action="store",dest="scale",\
                      metavar="INT",default=1.0,
                      help="grayscale contrast multiplier for points in plot")
    parser.add_option("-f","--flat",action="store_true",dest="flatten",\
                      default=False,help="show selected iterations in a single plot")

    if PY3:
        f = StringIO()
        parser.print_help(file=f)
        f.seek(0)
        if 'Options:' not in hypercube_measures.__doc__:
            hypercube_measures.__doc__ += NL+'Options:{0}'.format(f.read().split('Options:')[-1])
        f.close()
    else:
        if 'Options:' not in hypercube_measures.__doc__:
            hypercube_measures.__doc__ += NL+'Options:{0}'.format(parser.format_help().split('Options:')[-1])

    try:
        parsed_opts, parsed_args = parser.parse_args(cmdargs)
    except UnboundLocalError:
        pass
    if __quit: return

    # get the name of the parameter log file
    if instance is None:
        instance = parsed_args[0]
    from mystic.munge import read_history
    params, _cost = read_history(instance)
    # would be nice to use meta = ['wx','wx2','x','x2','wy',...]
    # exec "from {file} import meta".format(file=file)

    try: # path of plot output file
      out = parsed_opts.out  # e.g. 'myplot.png'
      if "None" == out: out = None
    except:
      out = None

    try: # select the bounds
        bounds = parsed_opts.bounds.split(",")  # format is "60:105, 0:30, 2.1:2.8"
        bounds = [tuple(float(j) for j in i.split(':')) for i in bounds]
    except:
        bounds = [(0,1),(0,1),(0,1)]

    try: # select which params are along which axes
        xyz = parsed_opts.xyz.split(",")  # format is "0 1, 4 5, 8 9"
        xyz = [tuple(int(j) for j in i.split()) for i in xyz]
    except:
        xyz = [(1,),(3,),(5,)]

    try: # select which params are along which axes
        wxyz = parsed_opts.wxyz.split(",")  # format is "0 1, 4 5, 8 9"
        wxyz = [tuple(int(j) for j in i.split()) for i in wxyz]
    except:
        wxyz = [(0,),(2,),(4,)]

    try: # select labels for the axes
        label = parsed_opts.label.split(',')  # format is "x, y, z"
    except:
        label = ['','','']

    x = params[max(xyz[0])]
    try: # select which iterations to plot
        select = parsed_opts.iters.split(',')  # format is "2, 4, 5, 6"
    except:
        select = ['-1']
       #select = ['0','1','2','3']

    try: # collapse non-consecutive iterations into a single plot...
        flatten = parsed_opts.flatten
    except:
        flatten = False

    try: # select which 'id' to plot results for
        id = int(parsed_opts.id)
    except:
        id = None # i.e. 'all' **or** use id=0, which should be 'best' energy ?

    try: # scale the color in plotting the weights
        scale = float(parsed_opts.scale)
    except:
        scale = 1.0 # color = color**scale

    # ensure all terms of bounds and xyz are tuples
    for bound in bounds:
        if not isinstance(bound, tuple):
            raise TypeError("bounds should be tuples of (lower_bound,upper_bound)")
    for i in range(len(xyz)):
        if isinstance(xyz[i], int):
            xyz[i] = (xyz[i],)
        elif not isinstance(xyz[i], tuple):
            raise TypeError("xyz should be tuples of (param1,param2,param3,...)")
    for i in range(len(wxyz)):
        if isinstance(wxyz[i], int):
            wxyz[i] = (wxyz[i],)
        elif not isinstance(wxyz[i], tuple):
            raise TypeError("wxyz should be tuples of (param1,param2,param3,...)")

    # ensure all terms of select are strings of ints
    for i in range(len(select)):
        if isinstance(select[i], int): select[i] = str(select[i])
        if select[i].count(':'):
            raise ValueError("iters should be ints")
       #if select[i] == '-1': select[i] = 'len(x)-1:len(x)'
       #elif not select[i].count(':'):
       #    select[i] += ':' + str(int(select[i])+1)

    # take only the selected 'id'
    if id != None:
        param = []
        for j in range(len(params)):
            param.append([p[id] for p in params[j]])
        params = param[:]

    # at this point, we should have:
    #bounds = [(60,105),(0,30),(2.1,2.8)] or [(None,None),(None,None),(None,None)]
    #xyz = [(2,3),(6,7),(10,11)] for any length tuple
    #wxyz = [(0,1),(4,5),(8,9)] for any length tuple (should match up with xyz)
    #select = ['-1'] or ['1','2','3','-1'] or similar
    #id = 0 or None

    plots = len(select)
    if not flatten:
        dim1,dim2 = best_dimensions(plots)
    else: dim1,dim2 = 1,1

    # use the default bounds where not specified
    bounds = [list(i) for i in bounds]
    for i in range(len(bounds)):
        if bounds[i][0] is None: bounds[i][0] = 0
        if bounds[i][1] is None: bounds[i][1] = 1

    # correctly bound the first plot.  there must be at least one plot
    import matplotlib.pyplot as plt
    fig = plt.figure()
    try:
        Subplot3D = None
        ax1 = fig.add_subplot(dim1, dim2, 1, projection='3d')
        newsubplotcode = "ax = ax{0:d} = fig.add_subplot(dim1,dim2,{0:d}, projection='3d');"
    except:
        from mpl_toolkits.mplot3d import Axes3D as _Axes3D
        from matplotlib.axes import subplot_class_factory
        Subplot3D = subplot_class_factory(_Axes3D)
        ax1 = Subplot3D(fig, dim1,dim2,1)
        newsubplotcode = "ax = ax{0:d} = Subplot3D(fig, dim1,dim2,{0:d});"
    ax1.plot([bounds[0][0]],[bounds[1][0]],[bounds[2][0]])
    ax1.plot([bounds[0][1]],[bounds[1][1]],[bounds[2][1]])
    globals = {'plt':plt,'fig':fig,'dim1':dim1,'dim2':dim2,\
               'Subplot3D':Subplot3D,'bounds':bounds,'label':label}
    if not flatten:
        code = "plt.title('iterations[{0}]');".format(select[0])
    else: 
        code = "plt.title('iterations[*]');"
    code = compile(code, '<string>', 'exec')
    %(exec_string)s
    ax1.set_xlabel(label[0])
    ax1.set_ylabel(label[1])
    ax1.set_zlabel(label[2])
    a = [ax1]

    # set up additional plots
    if not flatten:
        for i in range(2, plots + 1):
            code = newsubplotcode.format(i)
            code += "ax.plot([bounds[0][0]],[bounds[1][0]],[bounds[2][0]]);"
            code += "ax.plot([bounds[0][1]],[bounds[1][1]],[bounds[2][1]]);"
            code += "plt.title('iterations[{0}]');".format(select[i - 1])
            code += "ax.set_xlabel(label[0]);"
            code += "ax.set_ylabel(label[1]);"
            code += "ax.set_zlabel(label[2]);"
            code = compile(code, '<string>', 'exec')
            %(exec_string)s
            a.append(globals['ax'])

    # turn each "n:m" in select to a list
   #_select = []
   #for sel in select:
   #    if sel[0] == ':': _select.append("0"+sel)
   #    else: _select.append(sel)
   #for i in range(len(_select)):
   #    if _select[i][-1] == ':': select[i] = _select[i]+str(len(x))
   #    else: select[i] = _select[i]
   #for i in range(len(select)):
   #    p = select[i].split(":")
   #    if p[0][0] == '-': p[0] = "len(x)"+p[0]
   #    if p[1][0] == '-': p[1] = "len(x)"+p[1]
   #    select[i] = p[0]+":"+p[1]
   #steps = [eval("range({0})".format(sel.replace(":",","))) for sel in select]
    locals = dict(x=x, params=params, wxyz=wxyz, xyz=xyz)
    steps = [eval("[int({0})]".format(sel), locals) for sel in select]

    # at this point, we should have:
    #xyz = [(2,3),(6,7),(10,11)] for any length tuple
    #wxyz = [(0,1),(4,5),(8,9)] for any length tuple (should match up with xyz)
    #steps = [[0],[1],[3],[8]] or similar
    if flatten:
        from mystic.tools import flatten
        steps = [list(flatten(steps))]

    # adjust for logarithmic scaling of intensity
    from numpy import e
    scale = e**(scale - 1.0)

    # dot color is based on a product of weights
    t = []
    for v in range(len(steps)):
        t.append([])
        for s in steps[v]:
            for i in eval("[params[q][{0}] for q in wxyz[0]]".format(s), locals):
                for j in eval("[params[q][{0}] for q in wxyz[1]]".format(s), locals):
                    for k in eval("[params[q][{0}] for q in wxyz[2]]".format(s), locals):
                        t[v].append([str((1.0 - i[q]*j[q]*k[q])**scale) for q in range(len(i))])
                        if float(t[v][-1][-1]) > 1.0 or float(t[v][-1][-1]) < 0.0:
                            raise ValueError("Weights must be in range 0-1. Check normalization and/or assignment.")

    # build all the plots
    for v in range(len(steps)):
        for s in steps[v]:
            u = 0
            for i in eval("[params[q][{0}] for q in xyz[0]]".format(s), locals):
                for j in eval("[params[q][{0}] for q in xyz[1]]".format(s), locals):
                    for k in eval("[params[q][{0}] for q in xyz[2]]".format(s), locals):
                        for q in range(len(t[v][u])):
                            a[v].plot([i[q]],[j[q]],[k[q]],marker='o',color=t[v][u][q],ms=10)
                        u += 1

    # process inputs
    if _out: out = _out
    if out: out = _out or out

    if not out:
        plt.show()
    elif out is True:
        return fig #XXX: better?: fig.axes if len(fig.axes) > 1 else ax
    else:
        fig.savefig(out)
''' % dict(exec_string=exec_string)


def_hypercube_scenario = '''
def hypercube_scenario(filename, datafile=None, **kwds):
    """
generate scenario support plots from file written with ``write_support_file``;
and generate legacy data and cones from a dataset file, if provided

Available from the command shell as::

    support_hypercube_scenario filename (datafile) [options]

or as a function call::

    mystic.support.hypercube_scenario(filename, datafile=None, **options)

Args:
    filename (str): name of the convergence logfile (e.g. ``paramlog.py``)
    datafile (str, default=None): name of the dataset file (e.g. ``data.txt``)

Returns:
    None

Notes:
    - The option *out* takes a string of the filepath for the generated plot.
      If ``out = True``, return the Figure object instead of generating a plot.
    - The options *bounds*, *dim*, and *iters* all take indicator strings.
      The bounds should be given as comma-separated slices. For example, using
      ``bounds = ".062:.125, 0:30, 2300:3200"`` will set lower and upper bounds
      for x to be (.062,.125), y to be (0,30), and z to be (2300,3200). If all
      bounds are to not be strictly enforced, append an asterisk ``*`` to the
      string. The dim (dimensions of the scenario) should comma-separated ints.      For example, ``dim = "1, 1, 2"`` will convert the params to a two-member
      3-D dataset. Iters accepts a string built from comma-separated array
      slices. Thus, ``iters = ":"`` will plot all iters in a single plot.
      Alternatively, ``iters = ":2, 2:"`` will split the iters into two plots,
      while ``iters = "0"`` will only plot the first iteration.
    - The option *label* takes comma-separated strings. Thus ``label = "x,y,"``
      will place 'x' on the x-axis, 'y' on the y-axis, and nothing on the
      z-axis. LaTeX, such as ``label = "$ h $, $ a$, $ v$"`` will label the
      axes with standard LaTeX math formatting. Note that the leading space is
      required, while a trailing space aligns the text with the axis instead of      the plot frame.
    - The option "filter" is used to select datapoints from a given dataset,
      and takes comma-separated ints.
    - A "mask" is given as comma-separated ints. When the mask has more than
      one int, the plot will be 2D.
    - The option "vertical" will plot the dataset values on the vertical axis;
      for 2D plots, cones are always plotted on the vertical axis.
    - The option *nid* takes an integer of the nth simultaneous points to plot.
    - The option *scale* takes an integer as a grayscale contrast multiplier.
    - The option *gap* takes an integer distance from cone center to vertex.
    - The option *data* takes a boolean, to plot legacy data, if provided.
    - The option *cones* takes a boolean, to plot cones, if provided.
    - The option *flat* takes a boolean, to plot results in a single plot.
"""
    import shlex
    try:
        basestring
        from StringIO import StringIO
    except NameError:
        basestring = str
        from io import StringIO
    global __quit
    __quit = False
    _out = False

    instance = None
    # handle the special case where list is provided by sys.argv
    if isinstance(filename, (list,tuple)) and not kwds:
        cmdargs = filename # (above is used by script to parse command line)
    elif isinstance(filename, basestring) and not kwds:
        cmdargs = shlex.split(filename)
    # 'everything else' is essentially the functional interface
    else:
        cmdargs = kwds.get('kwds', '')
        if not cmdargs:
            out = kwds.get('out', None)
            bounds = kwds.get('bounds', None)
            iters = kwds.get('iters', None)
            label = kwds.get('label', None)
            dim = kwds.get('dim', None)
            filter = kwds.get('filter', None)
            mask = kwds.get('mask', None)
            nid = kwds.get('nid', None)
            scale = kwds.get('scale', None)
            gap = kwds.get('gap', None)
            data = kwds.get('data', False)
            cones = kwds.get('cones', False)
            vertical = kwds.get('vertical', False)
            flat = kwds.get('flat', False)

            if isinstance(out, bool): _out, out = out, None

            # process "commandline" arguments
            cmdargs = ''
            cmdargs += '' if out is None else '--out={} '.format(out)
            cmdargs += '' if bounds is None else '--bounds="{}" '.format(bounds)
            cmdargs += '' if iters is None else '--iters="{}" '.format(iters)
            cmdargs += '' if label is None else '--label="{}" '.format(label)
            cmdargs += '' if dim is None else '--dim="{}" '.format(dim)
            cmdargs += '' if filter is None else '--filter="{}" '.format(filter)
            cmdargs += '' if mask is None else '--mask={} '.format(mask)
            cmdargs += '' if nid is None else '--nid={} '.format(nid)
            cmdargs += '' if scale is None else '--scale={} '.format(scale)
            cmdargs += '' if gap is None else '--gap={} '.format(gap)
            cmdargs += '' if data == False else '--data '
            cmdargs += '' if cones == False else '--cones '
            cmdargs += '' if vertical == False else '--vertical '
            cmdargs += '' if flat == False else '--flat '
        else:
            cmdargs = ' ' + cmdargs
        if isinstance(filename, basestring):
            cmdargs = filename.split() + shlex.split(cmdargs)
        else: # special case of passing in monitor instance
            instance = filename
            cmdargs = shlex.split(cmdargs)

    #XXX: note that 'argparse' is new as of python2.7
    from optparse import OptionParser
    def _exit(self, errno=None, msg=None):
        global __quit
        __quit = True
        if errno or msg:
            msg = msg.split(': error: ')[-1].strip()
            raise IOError(msg)
    OptionParser.exit = _exit

    parser = OptionParser(usage=hypercube_scenario.__doc__.split(NL2+'Options:')[0])
    parser.add_option("-u","--out",action="store",dest="out",\
                      metavar="STR",default=None,
                      help="filepath to save generated plot")
    parser.add_option("-b","--bounds",action="store",dest="bounds",\
                      metavar="STR",default="0:1, 0:1, 0:1",
                      help="indicator string to set hypercube bounds")
    parser.add_option("-i","--iters",action="store",dest="iters",\
                      metavar="STR",default="-1",
                      help="indicator string to select iterations to plot")
    parser.add_option("-l","--label",action="store",dest="label",\
                      metavar="STR",default=",,",
                      help="string to assign label to axis")
    parser.add_option("-d","--dim",action="store",dest="dim",\
                      metavar="STR",default="None",
                      help="indicator string set to scenario dimensions")
    parser.add_option("-p","--filter",action="store",dest="filter",\
                      metavar="STR",default="None",
                      help="filter to select datapoints to plot")
    parser.add_option("-m","--mask",action="store",dest="replace",\
                      metavar="INT",default=None,
                      help="id # of the coordinate axis not to be plotted")
    parser.add_option("-n","--nid",action="store",dest="id",\
                      metavar="INT",default=None,
                      help="id # of the nth simultaneous points to plot")
    parser.add_option("-s","--scale",action="store",dest="scale",\
                      metavar="INT",default=1.0,
                      help="grayscale contrast multiplier for points in plot")
    parser.add_option("-g","--gap",action="store",dest="gap",\
                      metavar="INT",default=0.0,
                      help="distance from double cone center to cone vertex")
    parser.add_option("-o","--data",action="store_true",dest="legacy",\
                      default=False,help="plot legacy data, if provided")
    parser.add_option("-v","--cones",action="store_true",dest="cones",\
                      default=False,help="plot cones, if provided")
    parser.add_option("-z","--vertical",action="store_true",dest="vertical",\
                      default=False,help="always plot values on vertical axis")
    parser.add_option("-f","--flat",action="store_true",dest="flatten",\
                      default=False,help="show selected iterations in a single plot")

    if PY3:
        f = StringIO()
        parser.print_help(file=f)
        f.seek(0)
        if 'Options:' not in hypercube_scenario.__doc__:
            hypercube_scenario.__doc__ += NL+'Options:{0}'.format(f.read().split('Options:')[-1])
        f.close()
    else:
        if 'Options:' not in hypercube_scenario.__doc__:
            hypercube_scenario.__doc__ += NL+'Options:{0}'.format(parser.format_help().split('Options:')[-1])

    try:
        parsed_opts, parsed_args = parser.parse_args(cmdargs)
    except UnboundLocalError:
        pass
    if __quit: return

    # get the name of the parameter log file
    if instance is None:
        instance = parsed_args[0]
    from mystic.munge import read_history
    params, _cost = read_history(instance)
    # would be nice to use meta = ['wx','wx2','x','x2','wy',...]
    # exec "from {file} import meta".format(file=file)

    from mystic.math.discrete import scenario
    from mystic.math.legacydata import dataset
    try: # select whether to plot the cones
        cones = parsed_opts.cones
    except:
        cones = False

    try: # select whether to plot the legacy data
        legacy = parsed_opts.legacy
    except:
        legacy = False

    try: # path of plot output file
      out = parsed_opts.out  # e.g. 'myplot.png'
      if "None" == out: out = None
    except:
      out = None

    try: # get dataset filter
        filter = parsed_opts.filter
        if "None" == filter: filter = None
        else: filter = [int(i) for i in filter.split(",")] # format is "1,5,9"
    except:
        filter = None

    try: # select the scenario dimensions
        npts = parsed_opts.dim
        if "None" == npts: # npts may have been logged
            from mystic.munge import read_import
            npts = read_import(parsed_args[0], 'npts')
        else: npts = tuple(int(i) for i in dim.split(",")) # format is "1,1,1"
    except:
        npts = (1,1,1) #XXX: better in parsed_args ?

    try: # get the name of the dataset file
        file = parsed_args[1]
        from mystic.math.legacydata import load_dataset
        data = load_dataset(file, filter)
    except:
#       raise IOError("please provide dataset file name")
        data = dataset()
        cones = False
        legacy = False

    try: # select the bounds
        _bounds = parsed_opts.bounds
        if _bounds[-1] == "*": #XXX: then bounds are NOT strictly enforced
            _bounds = _bounds[:-1]
            strict = False
        else:
            strict = True
        bounds = _bounds.split(",")  # format is "60:105, 0:30, 2.1:2.8"
        bounds = [tuple(float(j) for j in i.split(':')) for i in bounds]
    except:
        strict = True
        bounds = [(0,1),(0,1),(0,1)]

    try: # select labels for the axes
        label = parsed_opts.label.split(',')  # format is "x, y, z"
    except:
        label = ['','','']

    x = params[-1]
    try: # select which iterations to plot
        select = parsed_opts.iters.split(',')  # format is ":2, 2:4, 5, 6:"
    except:
        select = ['-1']
       #select = [':']
       #select = [':1']
       #select = [':2','2:']
       #select = [':1','1:2','2:3','3:']
       #select = ['0','1','2','3']

    try: # collapse non-consecutive iterations into a single plot...
        flatten = parsed_opts.flatten
    except:
        flatten = False

    try: # select which 'id' to plot results for
        id = int(parsed_opts.id)
    except:
        id = None # i.e. 'all' **or** use id=0, which should be 'best' energy ?

    try: # scale the color in plotting the weights
        scale = float(parsed_opts.scale)
    except:
        scale = 1.0 # color = color**scale

    try: # gap between double cone center and cone vertex
        gap = float(parsed_opts.gap)
    except:
        gap = 0.0

    _2D = False # if False, use 3D plots; if True, use 3D plots
    cs = None
    try: # select which axis to plot 'values' on  (3D plot)
        xs = int(parsed_opts.replace)
    except:
        try: # select which axes to mask (2D plot)
            xs = (int(i) for i in parsed_opts.replace.split(",")) # format is "1,2"
            xs = list(reversed(sorted(set(xs))))
            cs = int(xs[-1]) if xs[-1] != xs[0] else None
            xs = int(xs[0])
            xs,cs = cs,xs # cs will swap coord axes; xs will swap with value axis
            _2D = True #NOTE: always apply cs swap before xs swap
        except:
            xs = None # don't plot values; plot values on 'x' axis with xs = 0

    try: # always plot cones on vertical axis
        vertical_cones = parsed_opts.vertical
    except:
        vertical_cones = False
    if _2D: # always plot cones on vertical axis
        vertical_cones = True

    # ensure all terms of bounds are tuples
    for bound in bounds:
        if not isinstance(bound, tuple):
            raise TypeError("bounds should be tuples of (lower_bound,upper_bound)")

    # ensure all terms of select are strings that have a ":"
    for i in range(len(select)):
        if isinstance(select[i], int): select[i] = str(select[i])
        if select[i] == '-1': select[i] = 'len(x)-1:len(x)'
        elif not select[i].count(':'):
            select[i] += ':' + str(int(select[i])+1)

    # take only the selected 'id'
    if id != None:
        param = []
        for j in range(len(params)):
            param.append([p[id] for p in params[j]])
        params = param[:]

    # at this point, we should have:
    #bounds = [(60,105),(0,30),(2.1,2.8)] or [(None,None)]*3
    #select = ['-1:'] or [':'] or [':1','1:2','2:3','3:'] or similar
    #id = 0 or None

    # get dataset coords (and values) for selected axes
    if data:
        slope = _get_slope(data, xs, cs)
        coords = _get_coords(data, xs, cs)
        #print("bounds: {0}".format(bounds))
        #print("slope: {0}".format(slope))
        #print("coords: {0}".format(coords))
   #else:
   #    slope = []
   #    coords = []

    plots = len(select)
    if not flatten:
        dim1,dim2 = best_dimensions(plots)
    else: dim1,dim2 = 1,1

    # use the default bounds where not specified
    bounds = [list(i) for i in bounds]
    for i in range(len(bounds)):
        if bounds[i][0] is None: bounds[i][0] = 0
        if bounds[i][1] is None: bounds[i][1] = 1

    # swap the axes appropriately when plotting cones (when replacing an axis)
    if _2D and xs == 0:
        if data:
            slope[0],slope[1] = slope[1],slope[0]
            coords = [list(reversed(pt[:2]))+pt[2:] for pt in coords]
       #bounds[0],bounds[1] = bounds[1],bounds[0]
       #label[0],label[1] = label[1],label[0]
        axis = xs #None
    elif not _2D and vertical_cones and xs in range(len(bounds)):
        # adjust slope, bounds, and data so cone axis is last 
        if data:
            slope = swap(slope, xs) 
            coords = [swap(pt,xs) for pt in coords]
        bounds = swap(bounds, xs) 
        label = swap(label, xs)
        axis = None
    else:
        axis = xs

    # correctly bound the first plot.  there must be at least one plot
    import matplotlib.pyplot as plt
    fig = plt.figure()
    try:
        Subplot3D = None
        if _2D:
            ax1 = fig.add_subplot(dim1, dim2, 1)
            newsubplotcode = "ax{0:d} = ax = fig.add_subplot(dim1,dim2,{0:d});"
        else:
            ax1 = fig.add_subplot(dim1, dim2, 1, projection='3d')
            newsubplotcode = "ax{0:d} = ax = fig.add_subplot(dim1,dim2,{0:d}, projection='3d');"
    except:
        from mpl_toolkits.mplot3d import Axes3D as _Axes3D
        from matplotlib.axes import subplot_class_factory
        Subplot3D = subplot_class_factory(_Axes3D)
        ax1 = Subplot3D(fig, dim1,dim2,1)
        newsubplotcode = "ax{0:d} = ax = Subplot3D(fig, dim1,dim2,{0:d});"
    if _2D:
        ax1.plot([bounds[0][0]],[bounds[1][0]])
        ax1.plot([bounds[0][1]],[bounds[1][1]])
    else:
        ax1.plot([bounds[0][0]],[bounds[1][0]],[bounds[2][0]])
        ax1.plot([bounds[0][1]],[bounds[1][1]],[bounds[2][1]])
    globals = {'plt':plt,'fig':fig,'dim1':dim1,'dim2':dim2,\
               'Subplot3D':Subplot3D,'bounds':bounds,'label':label}
    if not flatten:
        code = "plt.title('iterations[{0}]');".format(select[0])
    else: 
        code = "plt.title('iterations[*]');"
    code = compile(code, '<string>', 'exec')
    %(exec_string)s
    if cones and data and xs in range(len(bounds)):
        if _2D:
            _plot_bowtie(ax1,coords,slope,bounds,axis=axis,tol=gap)
        else:
            _plot_cones(ax1,coords,slope,bounds,axis=axis,strict=strict,tol=gap)
    if legacy and data:
        _plot_data(ax1,coords,bounds,strict=strict,_2D=_2D)
   #_clip_axes(ax1,bounds)
    _label_axes(ax1,label,_2D=_2D)
    a = [ax1]

    # set up additional plots
    if not flatten:
        for i in range(2, plots + 1):
            code = newsubplotcode.format(i)
            if _2D:
                code += "ax.plot([bounds[0][0]],[bounds[1][0]]);"
                code += "ax.plot([bounds[0][1]],[bounds[1][1]]);"
            else:
                code += "ax.plot([bounds[0][0]],[bounds[1][0]],[bounds[2][0]]);"
                code += "ax.plot([bounds[0][1]],[bounds[1][1]],[bounds[2][1]]);"
            code += "plt.title('iterations[{0}]');".format(select[i - 1])
            code = compile(code, '<string>', 'exec')
            %(exec_string)s
            ax = globals['ax']
            if cones and data and xs in range(len(bounds)):
                if _2D:
                    _plot_bowtie(ax,coords,slope,bounds,axis=axis,tol=gap)
                else:
                    _plot_cones(ax,coords,slope,bounds,axis=axis,strict=strict,tol=gap)
            if legacy and data:
                _plot_data(ax,coords,bounds,strict=strict,_2D=_2D)
           #_clip_axes(ax,bounds)
            _label_axes(ax,label,_2D=_2D)
            a.append(ax)

    # turn each "n:m" in select to a list
    _select = []
    for sel in select:
        if sel[0] == ':': _select.append("0"+sel)
        else: _select.append(sel)
    for i in range(len(_select)):
        if _select[i][-1] == ':': select[i] = _select[i]+str(len(x))
        else: select[i] = _select[i]
    for i in range(len(select)):
        p = select[i].split(":")
        if p[0][0] == '-': p[0] = "len(x)"+p[0]
        if p[1][0] == '-': p[1] = "len(x)"+p[1]
        select[i] = p[0]+":"+p[1]
    locals = dict(x=x, params=params)
    steps = [eval("range({0})".format(sel.replace(":",",")), locals) for sel in select]

    # at this point, we should have:
    #steps = [[0,1],[1,2],[2,3],[3,4,5,6,7,8]] or similar
    if flatten:
        from mystic.tools import flatten
        steps = [list(flatten(steps))]

    # plot all the scenario "data"
    from numpy import inf, e
    scale = e**(scale - 1.0)
    for v in range(len(steps)):
        if len(steps[v]) > 1: qp = float(max(steps[v]))
        else: qp = inf
        for s in steps[v]:
            par = eval("[params[q][{0}][0] for q in range(len(params))]".format(s), locals)
            pm = scenario()
            pm.load(par, npts)
            d = dataset()
            d.load(pm.coords, pm.values)
            # dot color determined by number of simultaneous iterations
            t = str((s/qp)**scale)
            # get and plot dataset coords for selected axes
            _coords = _get_coords(d, xs, cs)
            # check if we are replacing an axis
            if _2D and xs == 0:
                if data: # adjust data so cone axis is last 
                    _coords = [list(reversed(pt[:2]))+pt[2:] for pt in _coords]
            elif not _2D and vertical_cones and xs in range(len(bounds)):
                if data: # adjust data so cone axis is last 
                    _coords = [swap(pt,xs) for pt in _coords]
            _plot_data(a[v], _coords, bounds, color=t, strict=strict, _2D=_2D)

    if _2D: # strict = True
        plt.xlim((bounds[0][0],bounds[0][1]))
        plt.ylim((bounds[1][0],bounds[1][1]))

    # process inputs
    if _out: out = _out
    if out: out = _out or out

    if not out:
        plt.show()
    elif out is True:
        return fig #XXX: better?: fig.axes if len(fig.axes) > 1 else ax
    else:
        fig.savefig(out)
''' % dict(exec_string=exec_string)

exec(def_convergence)
exec(def_hypercube)
exec(def_hypercube_measures)
exec(def_hypercube_scenario)
del exec_string, def_convergence, def_hypercube
del def_hypercube_measures, def_hypercube_scenario


if __name__ == '__main__':
    pass


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Author: Alta Fang (altafang @caltech and alta @princeton)
# Copyright (c) 2008-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

# The following code attempts to construct something like:
# >>> from sympy import Eq, Symbol
# >>> from sympy import solve as symsol
# >>> x0 = Symbol('x0')
# >>> x1 = Symbol('x1')
# >>> x2 = Symbol('x2')
# >>> eq1 = Eq(x1, x0 - 2.)
# >>> eq2 = Eq(x1, x2*2.)
# >>> soln = symsol([eq2, eq1], [x0, x1, x2])

from mystic.tools import permutations
from mystic.tools import list_or_tuple_or_ndarray
import sys
if (sys.hexversion >= 0x30000f0):
    exec_locals_ = 'exec(code, _locals)'
    exec_globals_ = 'exec(_code, globals(), _locals)'
else:
    exec_locals_ = 'exec code in _locals'
    exec_globals_ = 'exec _code in globals(), _locals'
NL = '\n'
#FIXME: remove this head-standing to workaround python2.6 exec bug

def _classify_variables(constraints, variables='x', nvars=None): 
    """Takes a string of constraint equations and determines which variables
are dependent, independent, and unconstrained. Assumes there are no duplicate
equations. Returns a dictionary with keys: 'dependent', 'independent', and
'unconstrained', and with values that enumerate the variables that match
each variable type.

Inputs:
    constraints -- a string of symbolic constraints, with one constraint
        equation per line. Constraints must be equality constraints only.
        Standard python syntax should be followed (with the math and numpy
        modules already imported).

    For example:
        >>> constraints = '''
        ...     x0 = x4**2
        ...     x2 = x3 + x4'''
        >>> _classify_variables(constraints, nvars=5)
        {'dependent':['x0','x2'], 'independent':['x3','x4'], 'unconstrained':['x1']}
        >>> constraints = '''
        ...     x0 = x4**2
        ...     x4 - x3 = 0.
        ...     x4 - x0 = x2'''
        >>> _classify_variables(constraints, nvars=5)
        {'dependent': ['x0','x2','x4'], 'independent': ['x3'], 'unconstrained': ['x1']}

Additional Inputs:
    nvars -- number of variables. Includes variables not explicitly
        given by the constraint equations (e.g. 'x1' in the example above).
    variables -- desired variable name. Default is 'x'. A list of variable
        name strings is also accepted for when desired variable names
        don't have the same base, and can include variables that are not
        found in the constraints equation string.
"""
    if ">" in constraints or "<" in constraints:
        raise NotImplementedError("cannot classify inequalities") 

    from mystic.symbolic import replace_variables, get_variables
    #XXX: use solve? or first if not in form xi = ... ?
    if list_or_tuple_or_ndarray(variables):
        if nvars is not None: variables = variables[:nvars]
        constraints = replace_variables(constraints, variables)
        varname = '$'
        ndim = len(variables)
    else:
        varname = variables # varname used below instead of variables
        myvar = get_variables(constraints, variables)
        if myvar: ndim = max([int(v.strip(varname)) for v in myvar]) + 1
        else: ndim = 0
    if nvars is not None: ndim = nvars

    eqns = constraints.splitlines()
    indices = list(range(ndim))
    dep = []
    indep = []
    for eqn in eqns: # find which variables are used
        if eqn:
            for var in range(ndim):
                if indices.count(var) != 0:
                    if eqn.find(varname + str(var)) != -1:
                        indep.append(var)
                        indices.remove(var)
    indep.sort()
    _dep = []
    for eqn in eqns: # find which variables are on the LHS
        if eqn:
            split = eqn.split('=')
            for var in indep:
                if split[0].find(varname + str(var)) != -1:
                    _dep.append(var)
                    indep.remove(var)
                    break
    _dep.sort()
    indep = _dep + indep # prefer variables found on LHS
    for eqn in eqns: # find one dependent variable per equation
        _dep = []
        _indep = indep[:]
        if eqn:
            for var in _indep:
                if eqn.find(varname + str(var)) != -1:
                    _dep.append(var)
                    _indep.remove(var)
        if _dep:
            dep.append(_dep[0])
            indep.remove(_dep[0])
    #FIXME: 'equivalent' equations not ignored (e.g. x2=x2; or x2=1, 2*x2=2)
    """These are good:
    >>> constraints = '''
    ...     x0 = x4**2
    ...     x2 - x4 - x3 = 0.'''
    >>> _classify_variables(constraints, nvars=5)
    {'dependent': ['x0','x2'], 'independent': ['x3','x4'], 'unconstrained': ['x1']}
    >>> constraints = '''
    ...     x0 + x2 = 0.
    ...     x0 + 2*x2 = 0.'''
    >>> _classify_variables(constraints, nvars=5)
    {'dependent': ['x0','x2'], 'independent': [], 'unconstrained': ['x1','x3','x4']}

    This is a bug:
    >>> constraints = '''
    ...     x0 + x2 = 0.
    ...     2*x0 + 2*x2 = 0.'''
    >>> _classify_variables(constraints, nvars=5)
    {'dependent': ['x0','x2'], 'independent': [], 'unconstrained': ['x1','x3','x4']}
    """ #XXX: should simplify first?
    dep.sort()
    indep.sort()
    # return the actual variable names (not the indices)
    if varname == variables: # then was single variable
        variables = [varname+str(i) for i in range(ndim)]
    dep = [variables[i] for i in dep]
    indep = [variables[i] for i in indep]
    indices = [variables[i] for i in indices]
    d = {'dependent':dep, 'independent':indep, 'unconstrained':indices}
    return d


def _prepare_sympy(constraints, variables='x', nvars=None):
    """Parse an equation string and prepare input for sympy. Returns a tuple
of sympy-specific input: (code for variable declaration, left side of equation
string, right side of equation string, list of variables, and the number of
sympy equations).

Inputs:
    constraints -- a string of symbolic constraints, with one constraint
        equation per line. Constraints must be equality constraints only.
        Standard python syntax should be followed (with the math and numpy
        modules already imported).

    For example:
        >>> constraints = '''
        ...     x0 = x4**2
        ...     x4 - x3 = 0.
        ...     x4 - x0 = x2'''
        >>> code, lhs, rhs, vars, neqn = _prepare_sympy(constraints, nvars=5)
        >>> print(code)
        x0=Symbol('x0')
        x1=Symbol('x1')
        x2=Symbol('x2')
        x3=Symbol('x3')
        x4=Symbol('x4')
        rand = Symbol('rand')
        >>> print("%s %s" % (lhs, rhs))
        ['x0 ', 'x4 - x3 ', 'x4 - x0 '] [' x4**2', ' 0.', ' x2']
        >>> print("%s in %s eqns" % (vars, neqn))
        x0,x1,x2,x3,x4, in 3 eqns

Additional Inputs:
    nvars -- number of variables. Includes variables not explicitly
        given by the constraint equations (e.g. 'x1' in the example above).
    variables -- desired variable name. Default is 'x'. A list of variable
        name strings is also accepted for when desired variable names
        don't have the same base, and can include variables that are not
        found in the constraints equation string.
"""
    if ">" in constraints or "<" in constraints:
        raise NotImplementedError("cannot simplify inequalities") 

    from mystic.symbolic import replace_variables, get_variables
    #XXX: if constraints contain x0,x1,x3 for 'x', should x2 be in code,xlist?
    if list_or_tuple_or_ndarray(variables):
        if nvars is not None: variables = variables[:nvars]
        constraints = replace_variables(constraints, variables, markers='_')
        varname = '_'
        ndim = len(variables)
    else:
        varname = variables # varname used below instead of variables
        myvar = get_variables(constraints, variables)
        if myvar: ndim = max([int(v.strip(varname)) for v in myvar]) + 1
        else: ndim = 0
    if nvars is not None: ndim = nvars

    # split constraints_str into lists of left hand sides and right hand sides
    eacheqn = constraints.splitlines()
    neqns = 0
    left = []
    right = []
    for eq in eacheqn: #XXX: Le/Ge instead of Eq; Max/Min... (NotImplemented ?)
        splitlist = eq.replace('==','=').split('=') #FIXME: no inequalities
        if len(splitlist) == 2:   #FIXME: first convert >/< to min/max ?

            # If equation is blank on one side, raise error.
            if len(splitlist[0].strip()) == 0 or len(splitlist[1].strip()) == 0:
                print("%r is not an equation!" % eq) # Raise exception?
            else:
                left.append(splitlist[0])
                right.append(splitlist[1])
                neqns += 1

        # If equation doesn't have one equal sign, raise error.
        if len(splitlist) != 2 and len(splitlist) != 1:
            print("%r is not an equation!" % eq) # Raise exception?

    # First create list of x variables
    xlist = ""
    for i in range(ndim):
        xn = varname + str(i)
        xlist += xn + ","

    # Start constructing the code string
    code = ""
    for i in range(ndim):
        xn = varname + str(i)
        code += xn + '=' + "Symbol('" + xn + "')\n"

    code += "rand = Symbol('rand')\n"
    return code, left, right, xlist, neqns


def_solve_single = '''
def _solve_single(constraint, variables='x', target=None, **kwds):
    """Solve a symbolic constraints equation for a single variable.

Inputs:
    constraint -- a string of symbolic constraints. Only a single constraint
        equation should be provided, and must be an equality constraint. 
        Standard python syntax should be followed (with the math and numpy
        modules already imported).

    For example:
        >>> equation = "x1 - 3. = x0*x2"
        >>> print(_solve_single(equation))
        x0 = -(3.0 - x1)/x2

Additional Inputs:
    variables -- desired variable name. Default is 'x'. A list of variable
        name strings is also accepted for when desired variable names
        don't have the same base, and can include variables that are not
        found in the constraints equation string.
    target -- list providing the order for which the variables will be solved.
        By default, increasing order is used.

    For example:
        >>> equation = "x1 - 3. = x0*x2"
        >>> print(_solve_single(equation, target='x1'))
        x1 = 3.0 + x0*x2

Further Inputs:
    locals -- a dictionary of additional variables used in the symbolic
        constraints equations, and their desired values.
    simplest -- if True, simplify all but polynomials order >= 3 [False]
    rational -- if True, recast floats as rationals during solve [False]
    sequence -- if True, solve sequentially and not as a matrix [False]
    implicit -- if True, solve implicitly (with sin, cos, ...) [False]
    check -- if False, skip minimal testing (divide_by_zero, ...) [True]
    permute -- if True, return all permutations [False]
    warn -- if True, don't suppress warnings [True]
    verbose -- if True, print debug information [False]
""" #XXX: an very similar version of this code is found in _solve_linear XXX#
    # for now, we abort on multi-line equations or inequalities
    if len(constraint.replace('==','=').split('=')) != 2:
        raise NotImplementedError("requires a single valid equation") 
    if ">" in constraint or "<" in constraint:
        raise NotImplementedError("cannot simplify inequalities") 

    nvars = None
    permute = False # if True, return all permutations
    warn = True  # if True, don't suppress warnings
    verbose = False  # if True, print debug info
    simplest = False # if True, simplify all but polynomials order >= 3
    rational = False # if True, recast floats as rationals during solve
    sequence = False # if True, solve sequentially and not as a matrix
    implicit = False # if True, solve implicitly (contains sin, ...)
    check = True # if False, skip minimal testing (divide_by_zero, ...)
    kwarg = {} # keywords for sympy's symbolic solve

    permute = kwds['permute'] if 'permute' in kwds else permute
    warn = kwds['warn'] if 'warn' in kwds else warn
    verbose = kwds['verbose'] if 'verbose' in kwds else verbose
    kwarg['simplify'] = kwds['simplest'] if 'simplest' in kwds else simplest
    kwarg['rational'] = kwds['rational'] if 'rational' in kwds else rational
    kwarg['manual'] = kwds['sequence'] if 'sequence' in kwds else sequence
    kwarg['implicit'] = kwds['implicit'] if 'implicit' in kwds else implicit
    kwarg['check'] = kwds['check'] if 'check' in kwds else check

    if target in [None, False]:
        target = []
    elif isinstance(target, str):
        target = target.split(',')
    else:
        target = list(target) # not the best for ndarray, but should work

    from mystic.symbolic import replace_variables, get_variables
    if list_or_tuple_or_ndarray(variables):
        if nvars is not None: variables = variables[:nvars]
        constraints = replace_variables(constraint, variables, markers='_')
        varname = '_'
        ndim = len(variables)
        for i in range(len(target)):
            if variables.count(target[i]):
                target[i] = replace_variables(target[i],variables,markers='_')
    else:
        constraints = constraint # constraints used below
        varname = variables # varname used below instead of variables
        myvar = get_variables(constraint, variables)
        if myvar: ndim = max([int(v.strip(varname)) for v in myvar]) + 1
        else: ndim = 0
    if nvars is not None: ndim = nvars

    # create function to replace "_" with original variables
    def restore(variables, mystring):
        if list_or_tuple_or_ndarray(variables):
            vars = get_variables(mystring,'_')
            indices = [int(v.strip('_')) for v in vars]
            for i in range(len(vars)):
                mystring = mystring.replace(vars[i],variables[indices[i]])
        return mystring

    # default is _locals with sympy imported
    _locals = {}
    locals = kwds['locals'] if 'locals' in kwds else None
    if locals is None: locals = {}
    try:
        import imp
        imp.find_module('sympy')
        code = """from sympy import Eq, Symbol;"""
        code += """from sympy import solve as symsol;"""
        code = compile(code, '<string>', 'exec')
        %(exec_locals_)s
    except ImportError: # Equation will not be simplified."
        if warn: print("Warning: sympy not installed.")
        return constraint

    # default is _locals with numpy and math imported
    # numpy throws an 'AttributeError', but math passes error to sympy
    code = """from numpy import *; from math import *;""" # prefer math
    code += """from numpy import mean as average;""" # use np.mean not average
    code += """from numpy import var as variance;""" # look like mystic.math
    code += """from numpy import ptp as spread;"""   # look like mystic.math
    code = compile(code, '<string>', 'exec')
    %(exec_locals_)s
    _locals.update(dict(symsol_kwds=kwarg))
    _locals.update(locals) #XXX: allow this?

    code,left,right,xlist,neqns = _prepare_sympy(constraints, varname, ndim)

    eqlist = ""
    for i in range(1, neqns+1):
        eqn = 'eq' + str(i)
        eqlist += eqn + ","
        code += eqn + '= Eq(' + left[i-1] + ',' + right[i-1].replace(' 0.00', ' .00').replace(' 0.01', ' .01').replace(' 0.02', ' .02').replace(' 0.03', ' .03').replace(' 0.04', ' .04').replace(' 0.05', ' .05').replace(' 0.06', '.06').replace(' 0.07', ' .07').replace(' 0.08', ' .08').replace(' 0.09', ' .09').replace(' 0.0',' 0') + ')'+NL # sympy bug for 0.0
    eqlist = eqlist.rstrip(',')

    # get full list of variables in 'targeted' order
    xperms = xlist.split(',')[:-1]
    targeted = target[:]
    [targeted.remove(i) for i in targeted if i not in xperms]
    [targeted.append(i) for i in xperms if i not in targeted]
    _target = []
    [_target.append(i) for i in targeted if i not in _target]
    targeted = _target
    targeted = tuple(targeted)
    com = ',' if targeted else ''

    ########################################################################
    # solve each xi: symsol(single_equation, [x0,x1,...,xi,...,xn])
    # returns: {x0: f(xn,...), x1: f(xn,...), ..., xn: f(...,x0)}
    if permute or not target: #XXX: the goal is solving *only one* equation
        code += '_xlist = ({0}{1})'.format(','.join(targeted), com) + NL
        code += '_elist = [symsol(['+eqlist+'], [i], **symsol_kwds) for i in _xlist]' + NL
        code += '_elist = [i if isinstance(i, dict) else {j:i[-1][-1]} for j,i in zip(_xlist,_elist) if i]' + NL
        code += 'soln = dict()' + NL
        code += '[soln.update(i) for i in _elist if i]' + NL
    else:
        code += 'soln = symsol(['+eqlist+'], ['+target[0]+'], **symsol_kwds)' + NL
       #code += 'soln = symsol(['+eqlist+'], ['+targeted[0]+'], **symsol_kwds)' + NL
        code += 'soln = soln if isinstance(soln, dict) else {'+target[0]+': soln[-1][-1]} if soln else ""' + NL
    #code += 'print(soln)' + NL
    ########################################################################

    if verbose: print(code)
    _code = compile(code, '<string>', 'exec')
    try: 
        %(exec_globals_)s
        soln = _locals['soln'] if 'soln' in _locals else None
        if not soln:
            if warn: print("Warning: target variable is not valid")
            soln = {}
    except NotImplementedError: # catch 'multivariate' error for older sympy
        if warn: print("Warning: could not simplify equation.")
        return constraint      #FIXME: resolve diff with _solve_linear
    except NameError as error: # catch when variable is not defined
        if warn: print("Warning: {0}".format(error))
        soln = {}
    if verbose: print(soln)

    #XXX handles multiple solutions?
    soln = getattr(soln, 'iteritems', soln.items)()
    soln = dict([(str(key),str(value)) for key, value in soln])
    soln = [(i,soln[i]) for i in targeted if i in soln] #XXX: order as targeted?
    
    solns = []; solved = ""
    for key,value in soln:
        solved = str(key) + ' = ' + str(value) + NL
        if solved: solns.append( restore(variables, solved.rstrip()) )

    if not permute:
        sol = None if not solns[:1] else solns[0]
        return '' if sol and not sol.strip() else sol
    return tuple(solns)
''' % dict(exec_locals_=exec_locals_, exec_globals_=exec_globals_)
exec(def_solve_single)
del def_solve_single

doc_solve_linear = """Solve a system of symbolic linear constraints equations.

Inputs:
    constraints -- a string of symbolic constraints, with one constraint
        equation per line. Constraints must be equality constraints only.
        Standard python syntax should be followed (with the math and numpy
        modules already imported).

    For example:
        >>> constraints = '''
        ...     x0 - x2 = 2.
        ...     x2 = x3*2.'''
        >>> print(_solve_linear(constraints))
        x2 = 2.0*x3
        x0 = 2.0 + 2.0*x3

Additional Inputs:
    variables -- desired variable name. Default is 'x'. A list of variable
        name strings is also accepted for when desired variable names
        don't have the same base, and can include variables that are not
        found in the constraints equation string.
    target -- list providing the order for which the variables will be solved.
        If there are "N" constraint equations, the first "N" variables given
        will be selected as the dependent variables. By default, increasing
        order is used.

    For example:
        >>> constraints = '''
        ...     x0 - x2 = 2.
        ...     x2 = x3*2.'''
        >>> print(_solve_linear(constraints, target=['x3','x2']))
        x3 = -1.0 + 0.5*x0
        x2 = -2.0 + x0

Further Inputs:
    locals -- a dictionary of additional variables used in the symbolic
        constraints equations, and their desired values.
    simplest -- if True, simplify all but polynomials order >= 3 [False]
    rational -- if True, recast floats as rationals during solve [False]
    sequence -- if True, solve sequentially and not as a matrix [False]
    implicit -- if True, solve implicitly (with sin, cos, ...) [False]
    check -- if False, skip minimal testing (divide_by_zero, ...) [True]
    permute -- if True, return all permutations [False]
    warn -- if True, don't suppress warnings [True]
    verbose -- if True, print debug information [False]
"""
def_solve_linear = '''
def _solve_linear(constraints, variables='x', target=None, **kwds):
    nvars = None
    permute = False # if True, return all permutations
    warn = True  # if True, don't suppress warnings
    verbose = False  # if True, print debug info
    simplest = False # if True, simplify all but polynomials order >= 3
    rational = False # if True, recast floats as rationals during solve
    sequence = False # if True, solve sequentially and not as a matrix
    implicit = False # if True, solve implicitly (contains sin, ...)
    check = True # if False, skip minimal testing (divide_by_zero, ...)
    kwarg = {} # keywords for sympy's symbolic solve

    permute = kwds['permute'] if 'permute' in kwds else permute
    warn = kwds['warn'] if 'warn' in kwds else warn
    verbose = kwds['verbose'] if 'verbose' in kwds else verbose
    kwarg['simplify'] = kwds['simplest'] if 'simplest' in kwds else simplest
    kwarg['rational'] = kwds['rational'] if 'rational' in kwds else rational
    kwarg['manual'] = kwds['sequence'] if 'sequence' in kwds else sequence
    kwarg['implicit'] = kwds['implicit'] if 'implicit' in kwds else implicit
    kwarg['check'] = kwds['check'] if 'check' in kwds else check

    if target in [None, False]:
        target = []
    elif isinstance(target, str):
        target = target.split(',')
    else:
        target = list(target) # not the best for ndarray, but should work

    from mystic.symbolic import replace_variables, get_variables
    if list_or_tuple_or_ndarray(variables):
        if nvars is not None: variables = variables[:nvars]
        _constraints = replace_variables(constraints, variables, '_')
        varname = '_'
        ndim = len(variables)
        for i in range(len(target)):
            if variables.count(target[i]):
                target[i] = replace_variables(target[i],variables,markers='_')
    else:
        _constraints = constraints
        varname = variables # varname used below instead of variables
        myvar = get_variables(constraints, variables)
        if myvar: ndim = max([int(v.strip(varname)) for v in myvar]) + 1
        else: ndim = 0
    if nvars is not None: ndim = nvars

    # create function to replace "_" with original variables
    def restore(variables, mystring):
        if list_or_tuple_or_ndarray(variables):
            vars = get_variables(mystring,'_')
            indices = [int(v.strip('_')) for v in vars]
            for i in range(len(vars)):
                mystring = mystring.replace(vars[i],variables[indices[i]])
        return mystring

    # default is _locals with sympy imported
    _locals = {}
    locals = kwds['locals'] if 'locals' in kwds else None
    if locals is None: locals = {}
    # if sympy not installed, return original constraints
    try:
        import imp
        imp.find_module('sympy')
        code = """from sympy import Eq, Symbol;"""
        code += """from sympy import solve as symsol;"""
        code = compile(code, '<string>', 'exec')
        %(exec_locals_)s
    except ImportError: # Equation will not be simplified."
        if warn: print("Warning: sympy not installed.")
        return constraints

    # default is _locals with numpy and math imported
    # numpy throws an 'AttributeError', but math passes error to sympy
    code = """from numpy import *; from math import *;""" # prefer math
    code += """from numpy import mean as average;""" # use np.mean not average
    code += """from numpy import var as variance;""" # look like mystic.math
    code += """from numpy import ptp as spread;"""   # look like mystic.math
    code = compile(code, '<string>', 'exec')
    %(exec_locals_)s
    _locals.update(dict(symsol_kwds=kwarg))
    _locals.update(locals) #XXX: allow this?

    code,left,right,xlist,neqns = _prepare_sympy(_constraints, varname, ndim)

    eqlist = ""
    for i in range(1, neqns+1):
        eqn = 'eq' + str(i)
        eqlist += eqn + ","
        code += eqn + '= Eq(' + left[i-1] + ',' + right[i-1].replace(' 0.00', ' .00').replace(' 0.01', ' .01').replace(' 0.02', ' .02').replace(' 0.03', ' .03').replace(' 0.04', ' .04').replace(' 0.05', ' .05').replace(' 0.06', '.06').replace(' 0.07', ' .07').replace(' 0.08', ' .08').replace(' 0.09', ' .09').replace(' 0.0',' 0') + ')' + NL # sympy bug for 0.0
    eqlist = eqlist.rstrip(',')

    # get full list of variables in 'targeted' order
    xperms = xlist.split(',')[:-1]
    targeted = target[:]
    [targeted.remove(i) for i in targeted if i not in xperms]
    [targeted.append(i) for i in xperms if i not in targeted]
    _target = []
    [_target.append(i) for i in targeted if i not in _target]
    targeted = _target
    targeted = tuple(targeted)

    if permute:
        # Form constraints equations for each permutation.
        # This will change the order of the x variables passed to symsol()
        # to get different variables solved for.
        xperms = list(permutations(xperms)) #XXX: takes a while if nvars is ~10
        if target: # put the tuple with the 'targeted' order first
            xperms.remove(targeted)
            xperms.insert(0, targeted)
    else:
        xperms = [tuple(targeted)]

    solns = []
    for perm in xperms: 
        _code = code
        xlist = ','.join(perm).rstrip(',') #XXX: if not all, use target ?
        # solve dependent xi: symsol([linear_system], [x0,x1,...,xi,...,xn])
        # returns: {x0: f(xn,...), x1: f(xn,...), ...}
        _code += 'soln = symsol(['+eqlist+'], ['+xlist+'], **symsol_kwds)'
        #XXX: need to convert/check soln similarly as in _solve_single ?
        if verbose: print(_code)
        _code = compile(_code, '<string>', 'exec')
        try: 
            %(exec_globals_)s
            soln = _locals['soln'] if 'soln' in _locals else None
            if not soln:
                if warn: print("Warning: could not simplify equation.")
                soln = {}
        except NotImplementedError: # catch 'multivariate' error
            if warn: print("Warning: could not simplify equation.")
            soln = {}
        except NameError as error: # catch when variable is not defined
            if warn: print("Warning: {0}".format(error))
            soln = {}
        if verbose: print(soln)

        solved = ""
        for key, value in getattr(soln, 'iteritems', soln.items)():
            solved += str(key) + ' = ' + str(value) + NL
        if solved: solns.append( restore(variables, solved.rstrip()) )

    if not permute:
        return None if not solns[:1] else solns[0]

    # Remove duplicates
    filter = []; results = []
    for i in solns:
        _eqs = NL.join(sorted(i.split(NL)))
        if _eqs not in filter:
            filter.append(_eqs)
            results.append(i)
    return tuple(results)

_solve_linear.__doc__ = doc_solve_linear
''' % dict(exec_locals_=exec_locals_, exec_globals_=exec_globals_)
exec(def_solve_linear)
del def_solve_linear, doc_solve_linear, exec_locals_, exec_globals_

    # Create strings of all permutations of the solved equations.
    # Remove duplicates, then take permutations of the lines of equations
    # to create equations in different orders.
#   noduplicates = []
#   [noduplicates.append(i) for i in solns if i not in noduplicates]
#   stringperms = []
#   for item in noduplicates:
#       spl = item.splitlines()
#       for perm in permutations(spl):
#           permstring = ""
#           for line in perm:
#               permstring += line + '\n'
#           stringperms.append(permstring.rstrip())
#   return tuple(stringperms)


def solve(constraints, variables='x', target=None, **kwds):
    """Solve a system of symbolic constraints equations.

Inputs:
    constraints -- a string of symbolic constraints, with one constraint
        equation per line. Constraints must be equality constraints only.
        Standard python syntax should be followed (with the math and numpy
        modules already imported).

    For example:
        >>> constraints = '''
        ...     x0 - x2 = 2.
        ...     x2 = x3*2.'''
        >>> print(solve(constraints))
        x2 = 2.0*x3
        x0 = 2.0 + 2.0*x3
        >>> constraints = '''
        ...     spread([x0,x1]) - 1.0 = mean([x0,x1])   
        ...     mean([x0,x1,x2]) = x2'''
        >>> print(solve(constraints))
        x0 = -0.5 + 0.5*x2
        x1 = 0.5 + 1.5*x2

Additional Inputs:
    variables -- desired variable name. Default is 'x'. A list of variable
        name strings is also accepted for when desired variable names
        don't have the same base, and can include variables that are not
        found in the constraints equation string.
    target -- list providing the order for which the variables will be solved.
        If there are "N" constraint equations, the first "N" variables given
        will be selected as the dependent variables. By default, increasing
        order is used.

    For example:
        >>> constraints = '''
        ...     x0 - x2 = 2.
        ...     x2 = x3*2.'''
        >>> print(solve(constraints, target=['x3','x2']))
        x3 = -1.0 + 0.5*x0
        x2 = -2.0 + x0

Further Inputs:
    locals -- a dictionary of additional variables used in the symbolic
        constraints equations, and their desired values.
    simplest -- if True, simplify all but polynomials order >= 3 [False]
    rational -- if True, recast floats as rationals during solve [False]
    sequence -- if True, solve sequentially and not as a matrix [False]
    implicit -- if True, solve implicitly (with sin, cos, ...) [False]
    check -- if False, skip minimal testing (divide_by_zero, ...) [True]
    permute -- if True, return all permutations [False]
    warn -- if True, don't suppress warnings [False]
    verbose -- if True, print debug information [False]
"""
    kwds['warn'] = kwds.get('warn', False)
    try:
        if len(constraints.replace('==','=').split('=')) <= 2:
            soln = _solve_single(constraints, variables=variables, \
                                 target=target, **kwds)
            # for corner case where has something like: '0*xN'
            if not soln or not soln.strip():
                if target in [None, False]: target = []
                elif isinstance(target, str): target = target.split(',')
                else: target = list(target)
                soln = _solve_single(constraints, variables=variables, \
                                     target=target[1:], **kwds)
        else:
            soln = _solve_linear(constraints, variables=variables, \
                                 target=target, **kwds)
        if not soln: raise ValueError
    except:
        soln = _solve_nonlinear(constraints, variables=variables, \
                                target=target, **kwds)
    return soln


def _solve_nonlinear(constraints, variables='x', target=None, **kwds):
    """Build a constraints function given a string of nonlinear constraints.
Returns a constraints function. 

Inputs:
    constraints -- a string of symbolic constraints, with one constraint
        equation per line. Constraints must be equality constraints only.
        Standard python syntax should be followed (with the math and numpy
        modules already imported).

    For example:
        >>> constraints = '''x1 = x3*3. + x0*x2'''
        >>> print(_solve_nonlinear(constraints))
        x0 = (x1 - 3.0*x3)/x2
        >>> constraints = '''
        ...     spread([x0,x1]) - 1.0 = mean([x0,x1])   
        ...     mean([x0,x1,x2]) = x2'''
        >>> print(_solve_nonlinear(constraints))
        x0 = -0.5 + 0.5*x2
        x1 = 0.5 + 1.5*x2

Additional Inputs:
    variables -- desired variable name. Default is 'x'. A list of variable
        name strings is also accepted for when desired variable names
        don't have the same base, and can include variables that are not
        found in the constraints equation string.
    target -- list providing the order for which the variables will be solved.
        If there are "N" constraint equations, the first "N" variables given
        will be selected as the dependent variables. By default, increasing
        order is used.

    For example:
        >>> constraints = '''
        ...     spread([x0,x1]) - 1.0 = mean([x0,x1])   
        ...     mean([x0,x1,x2]) = x2'''
        >>> print(_solve_nonlinear(constraints, target=['x1']))
        x1 = -0.833333333333333 + 0.166666666666667*x2
        x0 = -0.5 + 0.5*x2

Further Inputs:
    locals -- a dictionary of additional variables used in the symbolic
        constraints equations, and their desired values.
    simplest -- if True, simplify all but polynomials order >= 3 [False]
    rational -- if True, recast floats as rationals during solve [False]
    sequence -- if True, solve sequentially and not as a matrix [False]
    implicit -- if True, solve implicitly (with sin, cos, ...) [False]
    check -- if False, skip minimal testing (divide_by_zero, ...) [True]
    permute -- if True, return all permutations [False]
    warn -- if True, don't suppress warnings [True]
    verbose -- if True, print debug information [False]
"""
    nvars = None
    permute = False # if True, return all permutations
    warn = True  # if True, don't suppress warnings
    verbose = False # if True, print details from _classify_variables
    simplest = False # if True, simplify all but polynomials order >= 3
    rational = False # if True, recast floats as rationals during solve
    sequence = False # if True, solve sequentially and not as a matrix
    implicit = False # if True, solve implicitly (contains sin, ...)
    check = True # if False, skip minimal testing (divide_by_zero, ...)
    kwarg = {} # keywords for sympy's symbolic solve

    permute = kwds['permute'] if 'permute' in kwds else permute
    warn = kwds['warn'] if 'warn' in kwds else warn
    verbose = kwds['verbose'] if 'verbose' in kwds else verbose
    kwarg['simplest'] = kwds['simplest'] if 'simplest' in kwds else simplest
    kwarg['rational'] = kwds['rational'] if 'rational' in kwds else rational
    kwarg['sequence'] = kwds['sequence'] if 'sequence' in kwds else sequence
    kwarg['implicit'] = kwds['implicit'] if 'implicit' in kwds else implicit
    kwarg['check'] = kwds['check'] if 'check' in kwds else check

    if target in [None, False]:
        target = []
    elif isinstance(target, str):
        target = target.split(',')
    else:
        target = list(target) # not the best for ndarray, but should work

    from mystic.symbolic import replace_variables, get_variables
    if list_or_tuple_or_ndarray(variables):
        if nvars is not None: variables = variables[:nvars]
        constraints = replace_variables(constraints, variables, '_')
        varname = '_'
        ndim = len(variables)
    else:
        varname = variables # varname used below instead of variables
        myvar = get_variables(constraints, variables)
        if myvar: ndim = max([int(v.strip(varname)) for v in myvar]) + 1
        else: ndim = 0
    if nvars is not None: ndim = nvars

    # create function to replace "_" with original variables
    def restore(variables, mystring):
        if list_or_tuple_or_ndarray(variables):
            vars = get_variables(mystring,'_')
            indices = [int(v.strip('_')) for v in vars]
            for i in range(len(vars)):
                mystring = mystring.replace(vars[i],variables[indices[i]])
        return mystring

    locals = kwds['locals'] if 'locals' in kwds else None
    if locals is None: locals = {}

    eqns = constraints.splitlines()
    # Remove empty strings:
    actual_eqns = []
    for j in range(len(eqns)):
        if eqns[j].strip():
           actual_eqns.append(eqns[j].strip())
    orig_eqns = actual_eqns[:]

    neqns = len(actual_eqns)

    xperms = [varname+str(i) for i in range(ndim)]
    if target:
        [target.remove(i) for i in target if i not in xperms]
        [target.append(i) for i in xperms if i not in target]
        _target = []
        [_target.append(i) for i in target if i not in _target]
        target = _target
    target = tuple(target)

    xperms = list(permutations(xperms)) #XXX: takes a while if nvars is ~10
    if target: # Try the suggested order first.
        xperms.remove(target)
        xperms.insert(0, target)

    complete_list = []

    constraints_function_list = []
    # Some of the permutations will give the same answer;
    # look into reducing the number of repeats?
    for perm in xperms:
        # Sort the list actual_eqns so any equation containing x0 is first, etc.
        sorted_eqns = []
        actual_eqns_copy = orig_eqns[:]
        usedvars = []
        for variable in perm: # range(ndim):
            for eqn in actual_eqns_copy:
                if eqn.find(variable) != -1:
                    sorted_eqns.append(eqn)
                    actual_eqns_copy.remove(eqn)
                    usedvars.append(variable)
                    break
        if actual_eqns_copy: # Append the remaining equations
            for item in actual_eqns_copy:
                sorted_eqns.append(item)
        actual_eqns = sorted_eqns

        # Append the remaining variables to usedvars
        tempusedvar = usedvars[:]
        tempusedvar.sort()
        nmissing = ndim - len(tempusedvar)
        for m in range(nmissing):
            usedvars.append(varname + str(len(tempusedvar) + m))

        #FIXME: not sure if the code below should be totally trusted...
        for i in range(neqns):
            # Trying to use xi as a pivot. Loop through the equations
            # looking for one containing xi.
            _target = usedvars[i%len(usedvars)] #XXX: ...to make it len of neqns
            for eqn in actual_eqns[i:]:
                invertedstring = _solve_single(eqn, variables=varname, target=_target, warn=warn, locals=locals, **kwarg)
                if invertedstring:
                    warn = False
                    break
            if invertedstring is None: continue #XXX: ...when _solve_single fails
            # substitute into the remaining equations. the equations' order
            # in the list newsystem is like in a linear coefficient matrix.
            newsystem = ['']*neqns
            j = actual_eqns.index(eqn)
            newsystem[j] = invertedstring #XXX: ...was eqn. I think correct now
            othereqns = actual_eqns[:j] + actual_eqns[j+1:]
            for othereqn in othereqns:
                expression = invertedstring.split("=")[1]
                fixed = othereqn.replace(_target, '(' + expression + ')')
                k = actual_eqns.index(othereqn)
                newsystem[k] = fixed
            actual_eqns = newsystem #XXX: potentially carrying too many eqns

        # Invert so that it can be fed properly to generate_constraint
        simplified = []
        for eqn in actual_eqns[:len(usedvars)]: #XXX: ...needs to be same len
            _target = usedvars[actual_eqns.index(eqn)]
            mysoln = _solve_single(eqn, variables=varname, target=_target, warn=warn, locals=locals, **kwarg)
            if mysoln: simplified.append(mysoln)
        simplified = restore(variables, '\n'.join(simplified).rstrip())

        if permute:
            complete_list.append(simplified)
            continue

        if verbose:
            print(_classify_variables(simplified, variables, ndim))
        return simplified

    warning='Warning: an error occurred in building the constraints.'
    if warn: print(warning)
    if verbose:
        print(_classify_variables(simplified, variables, ndim))
    if permute: #FIXME: target='x3,x1' may order correct, while 'x1,x3' doesn't
        filter = []; results = []
        for i in complete_list:
            _eqs = '\n'.join(sorted(i.split('\n')))
            if _eqs and (_eqs not in filter):
                filter.append(_eqs)
                results.append(i)
        return tuple(results) #FIXME: somehow 'rhs = xi' can be in results
    return simplified


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2018-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
plotter for data (x,z) and response surface function(*x)
  - initalize with x and z (and function)
  - interpolate if function is not provided
  - can downsample
  - plot data and response surface
"""

class Plotter(object):

    def __init__(self, x, z=None, function=None, **kwds):
        """scatter plotter for data (x,z) and response surface function(*x)

        Input:
          x: an array of shape (npts, dim) or (npts,)
          z: an array of shape (npts,) or (npts, N)
          function: function f, where z=f(*x.T), or str (interpolation method)

        Additional Inputs:
          step: int, plot every 'step' points on the grid [default: 200]
          scale: float, scaling factor for the z-axis [default: False]
          shift: float, additive shift for the z-axis [default: False]
          density: int, density of wireframe for the plot surface [default: 9]
          axes: tuple, indicies of the x-axes to plot [default: ()]
          axis: int, index of the z-axis to plot, if multi-dim [default: 0]
          vals: list of values (one per axis) for unplotted axes [default: ()]
          maxpts: int, maximum number of (x,z) points to use [default: None]
          kernel: function transforming x to x', where x' = kernel(x)
          vtol: float, maximum distance outside bounds hypercube to plot data

        NOTE:
          if scipy is not installed, will use np.interp for 1D (non-rbf),
          or mystic's rbf otherwise. default method is 'nearest' for
          1D and 'linear' otherwise. method can be one of ('rbf','linear',
          'nearest','cubic','inverse','gaussian','quintic','thin_plate').
        """
        self.x = getattr(x, '_x', x)  # params (x)
        self.z = x._y if z is None else z # cost (f(x))
        if function is None:
            function='linear'
        if type(function) is str:
            from mystic.math.interpolate import interpf #NOTE: axis may change
            function = interpf(self.x,self.z, method=function, arrays=True) #XXX: extrap, smooth, epsilon, norm?
        self.function = function
       #self.dim = kwds.pop('dim', None) #XXX: or len(x)?
        # interpolator configuration
        self.args = dict(step=200, scale=False, shift=False, vtol=None, \
            kernel=None, density=9, axes=(), vals=(), maxpts=None, axis=0)
        self.args.update(kwds)
        self.maxpts = self.args.pop('maxpts')
        return

    def _downsample(self, maxpts=None, x=None, z=None):
        """downsample (x,z) to at most maxpts

        Input:
          maxpts: int, maximum number of points to use from (x,z)
          x: an array of shape (npts, dim) or (npts,)
          z: an array of shape (npts,) or (npts, N)

        Output:
          x: an array of shape (npts, dim) or (npts,)
          z: an array of shape (npts,) or (npts, N)
        """
        if maxpts is None: maxpts = self.maxpts
        if x is None: x = self.x
        if z is None: z = self.z
        if len(x) != len(z):
            raise ValueError("the input array lengths must match exactly")
        if maxpts is not None and len(z) > maxpts:
            N = max(int(round(len(z)/float(maxpts))),1)
        #   print("for speed, sampling {} down to {}".format(len(z),len(z)/N))
        #   ax.plot(x[:,0], x[:,1], z, 'ko', linewidth=2, markersize=4)
            x = x[::N]
            z = z[::N]
        #   plt.show()
        #   exit()
        return x, z

    def _max(self, **kwds):
        """get the x[i],z[i] corresponding to the max(z)

        if z is multi-valued, accepts int axis to return single-valued z
        """
        import numpy as np
        mz = np.argmax(self.z, axis=0)
        x = np.asarray(self.x)[mz]
        z = np.asarray(self.z)[mz]
        zdim = len(z) if hasattr(z, '__len__') else 0
        if zdim > 1:
            axis = kwds.get('axis', self.args['axis']) # default = None?
            z = z.diagonal()
            if axis is not None:
                return (x[axis], z[axis])
        return (x,z)

    def _min(self, **kwds):
        """get the x[i],z[i] corresponding to the min(z)

        if z is multi-valued, accepts int axis to return single-valued z
        """
        import numpy as np
        mz = np.argmin(self.z, axis=0)
        x = np.asarray(self.x)[mz]
        z = np.asarray(self.z)[mz]
        zdim = len(z) if hasattr(z, '__len__') else 0
        if zdim > 1:
            axis = kwds.get('axis', self.args['axis']) # default = None?
            z = z.diagonal()
            if axis is not None:
                return (x[axis], z[axis])
        return (x,z)

    def Plot(self, **kwds):
        """produce a scatterplot of (x,z) and the surface z = function(*x.T)

        Input:
          step: int, plot every 'step' points on the grid [default: 200]
          scale: float, scaling factor for the z-axis [default: False]
          shift: float, additive shift for the z-axis [default: False]
          density: int, density of wireframe for the plot surface [default: 9]
          axes: tuple, indicies of the x-axes to plot [default: ()]
          axis: int, index of the z-axis to plot, if multi-dim [default: 0]
          vals: list of values (one per axis) for unplotted axes [default: ()]
          maxpts: int, maximum number of (x,z) points to use [default: None]
          kernel: function transforming x to x', where x' = kernel(x)
          vtol: float, maximum distance outside bounds hypercube to plot data
        """ #XXX: vtol can also be a tuple of vtols for each parameter
        step = kwds['step'] if 'step' in kwds else self.args['step']
        scale = kwds['scale'] if 'scale' in kwds else self.args['scale']
        shift = kwds['shift'] if 'shift' in kwds else self.args['shift']
        axes = kwds['axes'] if 'axes' in kwds else self.args['axes']
        axis = kwds['axis'] if 'axis' in kwds else self.args['axis']
        vals = kwds['vals'] if 'vals' in kwds else self.args['vals']
        maxpts = kwds['maxpts'] if 'maxpts' in kwds else self.maxpts
        kernel = kwds['kernel'] if 'kernel' in kwds else self.args['kernel']
        vtol = kwds['vtol'] if 'vtol' in kwds else self.args['vtol']
        density = kwds['density'] if 'density' in kwds else self.args['density']

        # plot response surface
        from mpl_toolkits.mplot3d import axes3d
        import matplotlib.pyplot as plt
        from matplotlib import cm
        import numpy as np
        from mystic.scripts import (_visual_filter as filtered,
                                    _parse_tol as parsetol)

        figure = plt.figure()
        kwds = {'projection':'3d'}
        ax = figure.axes[0] if figure.axes else plt.axes(**kwds)
        ax.autoscale(tight=True)

        zdim = len(self.z[0]) if hasattr(self.z[0], '__len__') else 0
        if zdim == 0:
            pass
        elif type(axis) is not int and zdim != 1:
            msg = "axis should be an int in the range 0 to %s" % (zdim-1)
            raise ValueError(msg)
        x, z = self._downsample(maxpts)
        x = np.asarray(x)
        z = np.asarray(z)
        z = z[:,axis] if zdim > 1 else z

        # get two axes to plot, and indices of the remaining axes
        axes = axes[:2]  #XXX: error if wrong size?
        ix = [i for i in range(len(x.T)) if i not in axes]
        n = 2-len(axes)
        axes, ix = list(axes)+ix[:n], ix[n:]

        # build list of fixed values (default mins), override with user input
       #fix = np.zeros(len(ix))
        fix = enumerate(self._min()[0])
        fix = np.array(tuple(j for (i,j) in fix if i not in axes))
        fix[:len(vals)] = vals

        # build grid of points, one for each param, apply fixed values
        grid = np.ones((len(x.T),step,step))
        grid[ix] = fix[:,None,None]

        # build sub-surface of function(x) to display, apply to the grid
        xy = x.T[axes]
        M = complex('{}j'.format(step))
        grid[axes] = np.mgrid[xy[0].min():xy[0].max():M,
                              xy[1].min():xy[1].max():M]

        # filter the plotted data points by the given vtol
        ixy = iter(xy)
        ifx = iter(fix)
        bounds = ""
        for i in range(max(axes + ix)+1):
            if i in axes:
                it = next(ixy)
                bounds += "%s:%s:%s" % (it.min(), it.max(), M)
            else:
                it = next(ifx)
                bounds += "%s" % it
            bounds += ", "
        bounds = bounds[:-2]
        del xy, fix, ix, ixy, ifx
        x, z = filtered(bounds, x, z, *parsetol(vtol, axes))

        # evaluate the function on the sub-surface
        z_ = np.asarray(self.function(*grid)) #XXX: is orientiation correct?
        z_ = z_[axis] if zdim > 1 else z_
        # scaling used by function plotter
        if scale:
            if shift:
                z_ = z_+shift
            z_ = np.log(4*z_*scale+1)+2

        # apply transform #NOTE: should do this w/o fixing points first
        if hasattr(kernel, '__call__'):
            _grid = np.zeros_like(grid[:2])
            for i in range(step):
                _grid.T[i] = [kernel(j)[:2] for j in grid.T[i]] #XXX: correct?
            grid = _grid
            ax0,ax1 = 0,1
        else: ax0,ax1 = axes

        # plot surface
        d = max(11 - density, 1)
        x_ = grid[ax0]
        y_ = grid[ax1]
        ax.plot_wireframe(x_, y_, z_, rstride=d, cstride=d, alpha=.3)
        #ax.plot_surface(x_, y_, z_, rstride=d, cstride=d, cmap=cm.jet, linewidth=0, antialiased=False)

        # use the sampled values
        z_ = np.asarray(z)
        # scaling used by function plotter
        if scale:
            if shift:
                z_ = z_+shift
            z_ = np.log(4*z_*scale+1)+2

        # apply transform
        if hasattr(kernel, '__call__'):
            x = np.array([kernel(j)[:2] for j in x])

        # plot data points
        x_ = x.T[ax0]
        y_ = x.T[ax1]
        ax.plot(x_, y_, z_, 'ko', linewidth=2, markersize=4)
        plt.show()  #XXX: show or don't show?... or return?


def plot(monitor, function=None, **kwds):
    '''generic interface to Plotter, returning an Plotter instance

    Input:
      monitor: a mystic.monitor instance
      function: function f, where z=f(*x.T), or str (interpolation method)

    Additional Inputs:
      step: int, plot every 'step' points on the grid [default: 200]
      scale: float, scaling factor for the z-axis [default: False]
      shift: float, additive shift for the z-axis [default: False]
      density: int, density of wireframe for the plot surface [default: 9]
      axes: tuple, indicies of the x-axes to plot [default: ()]
      axis: int, index of the z-axis to plot, if multi-dim [default: 0]
      vals: list of values (one per axis) for unplotted axes [default: ()]
      maxpts: int, maximum number of (x,z) points to use [default: None]
      kernel: function transforming x to x', where x' = kernel(x)
      vtol: float, maximum distance outside bounds hypercube to plot data

    NOTE:
      if scipy is not installed, will use np.interp for 1D (non-rbf),
      or mystic's rbf otherwise. default method is 'nearest' for
      1D and 'linear' otherwise. method can be one of ('rbf','linear',
      'nearest','cubic','inverse','gaussian','quintic','thin_plate').
    '''
    p = Plotter(monitor, function=function, **kwds)
    p.Plot()
    return p  #XXX: return nothing?


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2013-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
'''decorators for caching function outputs, with function inputs as the keys,
and interactors for reading and writing to databases of functions and data.

several klepto.cache strategies are included here for backward compatability;
please also see the klepto package for available archives (i.e. dir_archive,
file_archive, hdf_archive, sql_archive), cache strategies, filters, encodings,
serialization, and other configuration options
'''
from . import archive
from . import function

# backward compatability
from klepto import lru_cache, lfu_cache, mru_cache
from klepto import rr_cache, inf_cache, no_cache

def cached(**kwds):
    """build a caching archive for an objective function

    Input:
      type: the type of klepto.cache [default: inf_cache]
      archive: the archive (str name for a new archive, or archive instance)
      maxsize: maximum cache size [default: None]
      keymap: cache key encoder [default: klepto.keymap.keymap()]
      ignore: function argument names to ignore [default: '**']
      tol: int tolerance for rounding [default: None]
      deep: bool rounding depth (default: False]
      purge: bool for purging cache to archive [default: False]
      multivalued: bool if multivalued return of objective [default: False]

    Returns: 
      cached objective function

    Notes:
      inverse (y = -objective(x)) is at objective.__inverse__
      cache of objective is at objective.__cache__
      inverse and objective cache to the same archive
    """ #FIXME: ignore not handle *args as expected due to *args HACK
    _type = kwds.pop('type', inf_cache)
    multivalued = kwds.pop('multivalued', False)
    from klepto.keymaps import keymap as _keymap
    db = kwds.pop('archive', None)
    kwds.setdefault('keymap', _keymap())
    kwds.setdefault('ignore', '**') #FIXME: also ignores '*'
    if db is None: 
        kwds['cache'] = archive.read('archive')
    elif type(db) in (str, (u''.__class__)):
        kwds['cache'] = archive.read(db)
    else:
        kwds['cache'] = db
    # produce a cache with an archive backend
    cache = _type(**kwds)

    def dec(objective):
        """wrap a caching archive around an objective
        """
        # use a secretkey to store the objective's args
        SECRETKEY = '*' #XXX: too simple? #FIXME: ignore w/ignore='*', not names
        # wrap the cache around the objective function
        @cache
        def inner(*x, **kwds):
            args = kwds.pop(SECRETKEY, ())
            return objective(x, *args, **kwds)
        def _model(x, *args, **kwds):
            if args: kwds[SECRETKEY] = args
            return inner(*x, **kwds)
        _model.__inner__ = inner

        # when caching, always cache the multi-valued tuple
        if multivalued:
            def model(x, *argz, **kwdz):
                axis = kwdz.pop('axis', None)
                if axis is None: axis = slice(None)
                return _model(x, *argz, **kwdz)[axis]
        else:
            def model(x, *argz, **kwdz):
                axis = kwdz.pop('axis', None)
                return _model(x, *argz, **kwdz)
        # produce objective function that caches multi-valued output
        model.__cache__ = lambda : inner.__cache__()
        model.__doc__ = objective.__doc__

        # produce model inverse with shared cache
        imodel = lambda *args, **kwds: -model(*args, **kwds)
        model.__inverse__ = imodel
        imodel.__inverse__ = model
        imodel.__cache__ = model.__cache__

        return model
    return dec




#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Testing Rosenbrock's Function.

This is a very popular function for testing minimization algorithm.
The following provides tests for both bounded and unbounded minimization
of the Rosenbrock function with Differential Evolution.

For direct searches, Nelder-Mead does very well on this problem. 
For a direct comparison between DE and steepest-descent solvers, run
test_rosenbrock*.py (or optimize.py in scipy.optimize).
"""

from mystic.solvers import DifferentialEvolutionSolver, diffev
from mystic.termination import ChangeOverGeneration, VTR
from mystic.models import rosen
from mystic.monitors import Monitor, VerboseMonitor

from mystic.tools import random_seed
random_seed(123)

ND = 3
NP = 30
#MAX_GENERATIONS = 29
MAX_GENERATIONS = 99999

def main():
    solver = DifferentialEvolutionSolver(ND, NP)

    solver.SetRandomInitialPoints(min = [0]*ND, max = [2]*ND)
    solver.SetEvaluationLimits(generations=MAX_GENERATIONS)

    solver.Solve(rosen, termination = VTR(0.0001), \
                 CrossProbability=0.5, ScalingFactor=0.6, disp=1)

    solution = solver.bestSolution
   #print("Current function value: %s" % solver.bestEnergy)
   #print("Iterations: %s" % solver.generations)
   #print("Function evaluations: %s" % solver.evaluations)
  
    print(solution)



if __name__ == '__main__':
    from numpy import inf
    print("without bounds...")
    from timeit import Timer
    print("Differential Evolution")
    print("======================")
    t = Timer("main()", "from __main__ import main")
    timetaken =  t.timeit(number=1)
    print("CPU Time: %s\n" % timetaken)

    print("with bounds...")
    import time
    times = []
    algor = []

    print("Differential Evolution")
    print("======================")
    start = time.time()
    esow= Monitor()
    ssow= Monitor()
   #ssow= VerboseMonitor(1)

  # import random
 #  xinit = [random.random() for j in range(ND)]
    xinit = [0.8,1.2,0.7]
  # xinit = [0.8,1.2,1.7]             #... better when using "bad" range
    min = [-0.999, -0.999, 0.999]     #XXX: behaves badly when large range
    max = [200.001, 100.001, inf]     #... for >=1 x0 out of bounds; (up xtol)
  # min = [-0.999, -0.999, -0.999]
  # max = [200.001, 100.001, inf]
 #  min = [-0.999, -0.999, 0.999]     #XXX: tight range and non-randomness
 #  max = [2.001, 1.001, 1.001]       #...: is _bad_ for DE solvers

   #print(diffev(rosen,xinit,NP,retall=0,full_output=0))
    solver = DifferentialEvolutionSolver(len(xinit), NP)
    solver.SetInitialPoints(xinit)
    solver.SetStrictRanges(min,max)
    solver.SetEvaluationLimits(generations=MAX_GENERATIONS)
    solver.SetEvaluationMonitor(esow)
    solver.SetGenerationMonitor(ssow)
    solver.Solve(rosen, VTR(0.0001), \
                 CrossProbability=0.5, ScalingFactor=0.6, disp=1)
    sol = solver.bestSolution
    print(sol)
   #print("Current function value: %s" % solver.bestEnergy)
   #print("Iterations: %s" % solver.generations)
   #print("Function evaluations: %s" % solver.evaluations)
 
    times.append(time.time() - start)
    algor.append('Differential Evolution\t')

    for k,t in zip(algor,times):
        print("%s\t -- took %s" % (k, t))

   #print(len(esow.x))
   #print(len(ssow.x))
   #print("\nstep x:\n%s" % ssow.x[1:10][0])
   #print("\nstep y:\n%s" % ssow.y[1:10][0])



# end of file



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Tests functionality of misc. functions in scem.py
"""

from mystic.scemtools import *
import numpy

print("Numpy Input")

a = numpy.array([(i,i) for i in range(10)])*1.
c = [numpy.linalg.norm(x,2)  for x in a]

print("%s %s" % (a,c))
a, c = sort_complex(a,c)
print("%s %s" % (a,c))

print("List Input" )

a = numpy.array([(i,i) for i in range(10)])*1.
c = [numpy.linalg.norm(x,2)  for x in a]
a,c = list(a), list(c)
print("%s %s" % (a,c))
a, c = sort_complex(a,c)
print("%s %s" % (a,c))

print("update complex")
print("%s %s" % (a,c))
b = [2.5, 2.5]
d = 5.6
update_complex(a,c,b,d,0)
print("%s %s" % (a,c))

# end of file



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Input/output 'filters'
"""

# 'filters'
def identity(x):
    """identity filter, F, where F(x) yields x"""
    return x


def component(n, multiplier = 1.):
    """component filter, F, where F(x) yields x[n]"""
    def _(x):
        try:
            from numpy import asarray
            xx = asarray(x)
            xx = multiplier * xx[n,:]
            if isinstance(x[n],list):
                xx = xx.tolist()
            elif isinstance(x[n],tuple):
                xx = tuple(xx.tolist())
            elif type(x[n]) != type(xx):
                xx = type(x[n])(xx)
            return xx
        except IndexError:
            return multiplier * x[n]
    return _


# 'checkers'
def null_check(params, evalpts, *args):
    """null validity check"""
    return None


# monitor masks and filters
def generate_mask(cx=None, cy=None):
    """generate a data mask, based on constraints for x and/or y

    cx: mystic.constraint function where x' = cx(x) and x is parameter array
    cy: mystic.constraint function where [y'] = cy([y]) and y is cost array

    returns a masking function for (x,y) data or monitors, where
    list[bool] = mask(x,y), with True where constraints are satisfied

    For example:
    >>> mon = Monitor()
    >>> mon([0.0,0.5,1.0],2)
    >>> mon([2.0,3.0,4.0],3)
    >>> mon([4.0,5.0,6.0],4)
    >>> mon([5.0,5.5,6.5],6)
    >>> 
    >>> @impose_bounds((0,5))
    ... def inputs(x):
    ...   return x
    ... 
    >>> generate_mask(inputs)(mon)
    [True, True, False, False]
    >>> generate_mask(y=inputs)(mon)
    [True, True, True, False]
    >>>
    >>> @integers()
    ... def identity(x):
    ...   return x
    ... 
    >>> generate_mask(identity)(mon)
    [False, True, True, False]
    >>> generate_mask(y=identity)(mon)
    [True, True, True, True]
    """
    def func(x, z=None):
        _x = getattr(x, '_x', x)
        _y = x._y if z is None else z
        _x = _x.tolist() if hasattr(_x, 'tolist') else _x
        _y = _y.tolist() if hasattr(_y, 'tolist') else _y
        from numpy import ones_like
        res = ones_like(_y, dtype=bool)
        if cx is not None:
            res = res & [i == cx(i) for i in _x]
        if cy is not None:
            res = res & [(i,) == cy((i,)) for i in _y]
        return res.tolist()
    return func


def generate_filter(mask):
    """generate a data filter from a data masking function

    mask: masking function (built with generate_mask) or a boolean mask

    returns a function that filters (x,y) or a monitor, based on the given mask
    x',y' = filter(x,y), where filter removes values where mask is False

    For example:
    >>> mon = Monitor()
    >>> mon([0.0,0.5,1.0],2)
    >>> mon([2.0,3.0,4.0],3)
    >>> mon([4.0,5.0,6.0],4)
    >>> mon([5.0,5.5,6.5],6)
    >>> 
    >>> @impose_bounds((0,5))
    ... def inputs(x):
    ...   return x
    ... 
    >>> m = generate_filter(generate_mask(inputs))(mon)
    >>> m._x
    [[0.0, 0.5, 1.0], [2.0, 3.0, 4.0]]
    >>> m._y
    [2, 3]
    >>>
    >>> @integers()
    ... def identity(x):
    ...   return x
    ... 
    >>> m = generate_filter(generate_mask(identity))(mon)
    >>> m._x
    [[2.0, 3.0, 4.0], [4.0, 5.0, 6.0]]
    >>> m._y
    [3, 4]
    """
    def func(x, z=None):
        if not hasattr(mask, '__call__'):
            _mask = mask
        else:
            _mask = mask(x, z)
        if z is None and hasattr(x, '_x') and hasattr(x, '_y'):
            from copy import copy
            m = copy(x)
            mon = True
        else:
            from mystic.monitors import Monitor
            m = Monitor()
            m._x,m._y = x,z
            mon = False
        ax = True if hasattr(m._x, 'tolist') else False
        ay = True if hasattr(m._y, 'tolist') else False
        from numpy import asarray
        m._x = asarray(m._x)[_mask]
        m._y = asarray(m._y)[_mask]
        if not ax: m._x = m._x.tolist()
        if not ay: m._y = m._y.tolist()
        if mon:
            try:
                ai = True if hasattr(m._id, 'tolist') else False
                m._id = array(m._id)[_mask]
                if not ai: m._id = m._id.tolist()
            except IndexError:
                pass #XXX: m._id = []
            return m
        return m._x, m._y
    return func


# EOF
